# mypy: disable-error-code="attr-defined"
"""Test for the VectorNetworkAnalyzer E5080B class."""

import copy
from unittest.mock import MagicMock, patch

import pytest

from qililab.instrument_controllers.vector_network_analyzer.keysight_E5080B_vna_controller import E5080BController
from qililab.instruments.instrument import ParameterNotFound
from qililab.instruments.keysight.e5080b_vna import E5080B
from qililab.platform import Platform
from qililab.result.vna_result import VNAResult
from qililab.typings.enums import Parameter, VNAScatteringParameters, VNASweepModes, VNATriggerModes
from tests.data import SauronVNA
from tests.test_utils import build_platform


@pytest.fixture(name="sauron_platform")
def fixture_sauron_platform() -> Platform:
    """Return Platform object."""
    return build_platform(runcard=SauronVNA.runcard)


@pytest.fixture(name="e5080b_controller")
def fixture_e5080b_controller(sauron_platform: Platform):
    """Return an instance of VectorNetworkAnalyzer controller class"""
    settings = copy.deepcopy(SauronVNA.keysight_e5080b_controller)
    settings.pop("name")
    return E5080BController(settings=settings, loaded_instruments=sauron_platform.instruments)


@pytest.fixture(name="e5080b_no_device")
def fixture_e5080b_no_device():
    """Return an instance of VectorNetworkAnalyzer class"""
    settings = copy.deepcopy(SauronVNA.keysight_e5080b)
    settings.pop("name")
    return E5080B(settings=settings)


@pytest.fixture(name="e5080b")
@patch(
    "qililab.instrument_controllers.vector_network_analyzer.keysight_E5080B_vna_controller.VectorNetworkAnalyzerDriver",
    autospec=True,
)
def fixture_e5080b(mock_device: MagicMock, e5080b_controller: E5080BController):
    """Return connected instance of VectorNetworkAnalyzer class"""
    mock_instance = mock_device.return_value
    e5080b_controller.connect()
    mock_device.assert_called()
    return e5080b_controller.modules[0]


class TestE5080B:
    """Unit tests checking the VectorNetworkAnalyzer E5080B attributes and methods"""

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.POWER, -60.0),
            (Parameter.FREQUENCY_SPAN, 6.4e-3),
            (Parameter.FREQUENCY_CENTER, 8.5e-3),
            (Parameter.FREQUENCY_START, 27.5),
            (Parameter.FREQUENCY_STOP, 40.5),
            (Parameter.IF_BANDWIDTH, 50.0),
            (Parameter.DEVICE_TIMEOUT, 100.0),
            (Parameter.ELECTRICAL_DELAY, 0.0),
        ],
    )
    def test_setup_method_value_flt(self, parameter: Parameter, value, e5080b: E5080B):
        """Test the setup method with float value"""
        assert isinstance(parameter, Parameter)
        assert isinstance(value, float)
        e5080b.setup(parameter, value)
        if parameter == Parameter.POWER:
            assert e5080b.settings.power == value
        if parameter == Parameter.FREQUENCY_SPAN:
            assert e5080b.frequency_span == value
        if parameter == Parameter.FREQUENCY_CENTER:
            assert e5080b.frequency_center == value
        if parameter == Parameter.FREQUENCY_START:
            assert e5080b.frequency_start == value
        if parameter == Parameter.FREQUENCY_STOP:
            assert e5080b.frequency_stop == value
        if parameter == Parameter.IF_BANDWIDTH:
            assert e5080b.settings.if_bandwidth == value
        if parameter == Parameter.DEVICE_TIMEOUT:
            assert e5080b.device_timeout == value
        if parameter == Parameter.ELECTRICAL_DELAY:
            assert e5080b.electrical_delay == value

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.CURRENT, 0.34),
            (Parameter.VOLTAGE, -20.1),
        ],
    )
    def test_setup_method_flt_raises_exception(self, parameter, value, e5080b: E5080B):
        """Test the setup method raises exception with incorrect float parameter"""
        with pytest.raises(ParameterNotFound):
            e5080b.setup(parameter, value)

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.TRIGGER_MODE, "INT"),
            (Parameter.SCATTERING_PARAMETER, "S21"),
            (Parameter.SWEEP_MODE, VNASweepModes.CONT),
        ],
    )
    def test_setup_method_value_str(self, parameter: Parameter, value, e5080b: E5080B):
        """Test the setup method with str value"""
        assert isinstance(parameter, Parameter)
        assert isinstance(value, str)
        e5080b.setup(parameter, value)
        if parameter == Parameter.TRIGGER_MODE:
            assert e5080b.trigger_mode == VNATriggerModes(value)
        if parameter == Parameter.SCATTERING_PARAMETER:
            assert e5080b.scattering_parameter == VNAScatteringParameters(value)
        if parameter == Parameter.SWEEP_MODE:
            assert e5080b.get_sweep_mode() == VNASweepModes(value)

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.SCATTERING_PARAMETER, "S221"),
            (Parameter.SCATTERING_PARAMETER, "s11"),
            (Parameter.SWEEP_MODE, "CONT"),
            (Parameter.SWEEP_MODE, "sweep_mode continuous"),
        ],
    )
    def test_setup_method_value_str_raises_exception(self, parameter: Parameter, value, e5080b: E5080B):
        """Test the setup method raises exception with incorrect str value"""
        assert isinstance(parameter, Parameter)
        assert isinstance(value, str)
        with pytest.raises(ValueError):
            e5080b.setup(parameter, value)

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.CURRENT, "foo"),
            (Parameter.VOLTAGE, "bar"),
        ],
    )
    def test_setup_method_str_raises_exception(self, parameter, value, e5080b: E5080B):
        """Test the setup method raises exception with incorrect str parameter"""
        with pytest.raises(ParameterNotFound):
            e5080b.setup(parameter, value)

    @pytest.mark.parametrize(
        "parameter, value", [(Parameter.AVERAGING_ENABLED, True), (Parameter.AVERAGING_ENABLED, False)]
    )
    def test_setup_method_value_bool(self, parameter: Parameter, value, e5080b: E5080B):
        """Test the setup method with bool value"""
        assert isinstance(parameter, Parameter)
        assert isinstance(value, bool)
        e5080b.setup(parameter, value)
        if parameter == Parameter.AVERAGING_ENABLED:
            assert e5080b.averaging_enabled == value

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.CURRENT, True),
            (Parameter.VOLTAGE, False),
        ],
    )
    def test_setup_method_bool_raises_exception(self, parameter, value, e5080b: E5080B):
        """Test the setup method raises exception with incorrect bool parameter"""
        with pytest.raises(ParameterNotFound):
            e5080b.setup(parameter, value)

    @pytest.mark.parametrize("parameter, value", [(Parameter.NUMBER_POINTS, 100), (Parameter.NUMBER_AVERAGES, 4)])
    def test_setup_method_value_int(self, parameter: Parameter, value, e5080b: E5080B):
        """Test the setup method with int value"""
        assert isinstance(parameter, Parameter)
        assert isinstance(value, int)
        e5080b.setup(parameter, value)
        if parameter == Parameter.NUMBER_POINTS:
            assert e5080b.number_points == value
        if parameter == Parameter.NUMBER_AVERAGES:
            assert e5080b.number_averages == value

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.CURRENT, 0),
            (Parameter.VOLTAGE, -20),
        ],
    )
    def test_setup_method_int_raises_exception(self, parameter, value, e5080b: E5080B):
        """Test the setup method raises exception with incorrect int parameter"""
        with pytest.raises(ParameterNotFound):
            e5080b.setup(parameter, value)

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.SCATTERING_PARAMETER, ["S221"]),
            (Parameter.SCATTERING_PARAMETER, {}),
        ],
    )
    def test_setup_method_raises_exception(self, parameter, value, e5080b: E5080B):
        """Test the setup method raises exception with incorrect value type"""
        with pytest.raises(ParameterNotFound):
            e5080b.setup(parameter, value)

    @pytest.mark.parametrize(
        "parameter, value",
        [
            # Test POWER setting
            (Parameter.POWER, 2.0),

            # Test IF_BANDWIDTH setting
            (Parameter.IF_BANDWIDTH, 1.0),

            # Test ELECTRICAL_DELAY setting
            (Parameter.ELECTRICAL_DELAY, 1.0),

            # Test SWEEP_MODE setting
            (Parameter.SWEEP_MODE, VNASweepModes.HOLD),

            # Test DEVICE_TIMEOUT setting
            (Parameter.DEVICE_TIMEOUT, 1.0)
        ]
    )
    def test_set_parameter(self, e5080b: E5080B, parameter, value):
        """Test setting parameters for E5071B functionality using parameterized values."""
        e5080b.set_parameter(parameter, value, channel_id=1)

        # Check values based on the parameter
        if parameter == Parameter.POWER:
            assert e5080b.settings.power == value
        elif parameter == Parameter.IF_BANDWIDTH:
            assert e5080b.settings.if_bandwidth == value
        elif parameter == Parameter.ELECTRICAL_DELAY:
            assert e5080b.settings.electrical_delay == value
        elif parameter == Parameter.SWEEP_MODE:
            assert e5080b.settings.sweep_mode == value
        elif parameter == Parameter.DEVICE_TIMEOUT:
            assert e5080b.settings.device_timeout == value

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.GATE_DURATION, 0),
        ],
    )
    def test_set_parameter_raises_exception(self, parameter, value, e5080b: E5080B):
        """Test the setup method raises exception with not supported parameter"""
        with pytest.raises(ParameterNotFound):
            e5080b.set_parameter(parameter, value)

    @pytest.mark.parametrize(
        "parameter, expected_value, channel_id",
        [
            # Test POWER setting
            (Parameter.POWER, -60.0, None),
            # Test IF_BANDWITH setting
            (Parameter.IF_BANDWIDTH, 1.5, None),
            # Test ELECTRICAL_DELAY setting
            (Parameter.ELECTRICAL_DELAY, 0.0, None),
            # Test SWEEP_MODE setting with channel
            (Parameter.SWEEP_MODE, VNASweepModes.HOLD, 1),
            # Test SWEEP_MODE setting without channel
            (Parameter.SWEEP_MODE, VNASweepModes.HOLD, None),
            # Test DEVICE_TIMEOUT setting
            (Parameter.DEVICE_TIMEOUT, 10000, None)
        ]
    )
    @patch("qililab.instruments.keysight.e5080b_vna.E5080B._get_sweep_mode")
    def test_get_parameter(self, mock_get_sweep_mode, e5080b: E5080B, parameter, expected_value, channel_id):
        """Test setting parameters for E5071B functionality using parameterized values."""
        mock_get_sweep_mode.return_value = VNASweepModes.HOLD.name
        if parameter == Parameter.IF_BANDWIDTH:
            e5080b.set_parameter(Parameter.IF_BANDWIDTH, 1.5)
        value = e5080b.get_parameter(parameter, channel_id=channel_id)
        # Assert _get_sweep_mode was called correctly
        if parameter == Parameter.SWEEP_MODE:
            if channel_id:
                mock_get_sweep_mode.assert_called_once_with(channel=channel_id)
            else:
                mock_get_sweep_mode.assert_called_once_with()
            assert value == expected_value.name
        else:
            assert value == expected_value

    @pytest.mark.parametrize(
        "parameter, value",
        [
            (Parameter.GATE_DURATION, 0),
        ],
    )
    def test_get_parameter_raises_exception(self, parameter, value, e5080b: E5080B):
        """Test the setup method raises exception with not supported parameter"""
        with pytest.raises(ParameterNotFound):
            e5080b.get_parameter(parameter)

    def test_to_dict_method(self, e5080b_no_device: E5080B):
        """Test the dict method"""
        assert isinstance(e5080b_no_device.to_dict(), dict)

    def test_initial_setup_method(self, e5080b: E5080B):
        """Test the initial setup method"""
        e5080b.initial_setup()
        e5080b.device.initial_setup.assert_called()

    def test_reset_method(self, e5080b: E5080B):
        """Test the reset method"""
        e5080b.reset()
        e5080b.device.reset.assert_called()

    def test_turn_on_method(self, e5080b: E5080B):
        """Test turn on method"""
        e5080b.turn_on()
        e5080b.device.send_command.assert_called_with(command=":OUTP", arg="ON")

    def test_turn_off_method(self, e5080b: E5080B):
        """Test turn off method"""
        e5080b.turn_off()
        e5080b.device.send_command.assert_called_with(command=":OUTP", arg="OFF")

    @pytest.mark.parametrize("command", [":SENS1:AVER:CLE", "SENS1:AVER:COUN 3"])
    def test_send_command_method(self, command: str, e5080b: E5080B):
        """Test the send command method"""
        assert isinstance(command, str)
        e5080b.send_command(command)
        e5080b.device.send_command.assert_called_with(command=command, arg="?")

    @pytest.mark.parametrize("query", [":SENS1:SWE:MODE?"])
    def test_send_query_method(self, query: str, e5080b: E5080B):
        """Test the send query method"""
        assert isinstance(query, str)
        e5080b.send_query(query)
        e5080b.device.send_query.assert_called_with(query)

    @pytest.mark.parametrize("query", ["SENS:X?", "CALC1:MEAS1:DATA:SDAT?"])
    def test_send_binary_query_method(self, query: str, e5080b: E5080B):
        """Test the send binary query method"""
        assert isinstance(query, str)
        e5080b.send_binary_query(query)
        e5080b.device.send_binary_query.assert_called_with(query)

    def test_read_method(self, e5080b: E5080B):
        """Test the read method"""
        e5080b.read()
        e5080b.device.read.assert_called()

    def test_read_raw_method(self, e5080b: E5080B):
        """Test the read method"""
        e5080b.read_raw()
        e5080b.device.read_raw.assert_called()

    def test_set_timeout_method(self, e5080b: E5080B):
        """Test the set timeout method"""
        e5080b.set_timeout(100)
        e5080b.device.set_timeout.assert_called_with(100)

    @pytest.mark.parametrize(
        "state, command, arg", [(True, "SENS1:AVER:STAT", "ON"), (False, "SENS1:AVER:STAT", "OFF")]
    )
    def test_average_state_method(self, state, command, arg, e5080b: E5080B):
        """Test the auxiliary private method average state"""
        e5080b._average_state(state)
        e5080b.device.send_command.assert_called_with(command=command, arg=arg)

    def test_average_count_method(self, e5080b: E5080B):
        """Set the average count"""
        e5080b._average_count(1, 1)  # Have to know if you can select which one was called
        e5080b.device.send_command.assert_called()

    def test_autoscale_method(self, e5080b: E5080B):
        """Test the autoscale method"""
        e5080b.autoscale()
        e5080b.device.send_command.assert_called_with(command="DISP:WIND:TRAC:Y:AUTO", arg="")

    def test_get_sweep_mode_method(self, e5080b: E5080B):
        """Test the get sweep mode method"""
        output = e5080b._get_sweep_mode()
        e5080b.device.send_query.assert_called()
        assert isinstance(output, str)

    def test_get_trace_method(self, e5080b: E5080B):
        """Test auxiliarty private method get trace."""
        e5080b._get_trace()
        e5080b.device.send_command.assert_called()
        e5080b.device.send_binary_query.assert_called_once()

    @pytest.mark.parametrize("count", ["1", "3", "5"])
    def test_set_count_method(self, count: str, e5080b: E5080B):
        """Test the auxiliary private method set count"""
        e5080b._set_count(count)
        e5080b.device.send_command.assert_called_with(command="SENS1:SWE:GRO:COUN", arg=count)

    def test_pre_measurement_method(self, e5080b: E5080B):
        """Test the auxiliary private method pre measurment"""
        e5080b._pre_measurement()
        assert e5080b.averaging_enabled

    def test_start_measurement_method(self, e5080b: E5080B):
        """Test the auxiliary private method start measurment"""
        e5080b._start_measurement()
        assert e5080b.get_sweep_mode() == VNASweepModes("group")

    @patch("qililab.instruments.keysight.e5080b_vna.E5080B.ready")
    def test_wait_until_ready_method(self, mock_ready, e5080b: E5080B):
        """Test the auxiliary private method wait until ready"""
        mock_ready.return_value = True
        output = e5080b._wait_until_ready()
        assert output

    @patch("qililab.instruments.keysight.e5080b_vna.E5080B.ready")
    @patch.object(E5080B, "device_timeout", new=10)
    def test_wait_until_ready_method_fails(self, mock_ready, e5080b: E5080B):
        """Test the auxiliary private method wait until ready fails"""
        mock_ready.return_value = False
        output = e5080b._wait_until_ready()
        assert not output

    def test_average_clear_method(self, e5080b: E5080B):
        """Test the average clear method"""
        e5080b.average_clear()
        e5080b.device.send_command.assert_called()

    def test_get_frequencies_method(self, e5080b: E5080B):
        """Test the get frequencies method"""
        e5080b.get_frequencies()
        e5080b.device.send_binary_query.assert_called()

    def test_ready_method(self, e5080b: E5080B):
        """Test ready method"""
        output = e5080b.ready()
        assert isinstance(output, bool)

    @patch("qililab.instruments.keysight.e5080b_vna.E5080B._get_sweep_mode")
    def test_ready_method_raises_exception(self, mock_get_sweep_mode, e5080b: E5080B):
        """Test read method raises an Exception"""
        mock_get_sweep_mode.side_effect = ValueError("Mocked exception")
        output = e5080b.ready()
        assert not output
        with pytest.raises(Exception) as exc:
            e5080b.ready()
            assert exc is Exception

    def test_release_method(self, e5080b: E5080B):
        """Test release method"""
        e5080b.release()
        assert e5080b.settings.sweep_mode == VNASweepModes("cont")

    @patch("qililab.instruments.keysight.e5080b_vna.E5080B.ready")
    def test_read_tracedata_method(self, mock_ready, e5080b: E5080B):
        """Test the read tracedata method"""
        mock_ready.return_value = True
        output = e5080b.read_tracedata()
        assert output is not None

    @patch("qililab.instruments.keysight.e5080b_vna.E5080B.ready")
    @patch.object(E5080B, "device_timeout", new=10)
    def test_read_tracedata_method_raises_exception(self, mock_ready, e5080b: E5080B):
        """Test the read tracedata method"""
        mock_ready.return_value = False
        with pytest.raises(TimeoutError):
            e5080b.read_tracedata()

    @patch("qililab.instruments.keysight.e5080b_vna.E5080B.ready")
    def test_acquire_result_method(self, mock_ready, e5080b: E5080B):
        """Test the acquire result method"""
        mock_ready.return_value = True
        output = e5080b.acquire_result()
        assert isinstance(output, VNAResult)

    def test_scattering_parameter_property(self, e5080b_no_device: E5080B):
        """Test the scattering parametter property"""
        assert hasattr(e5080b_no_device, "scattering_parameter")
        assert e5080b_no_device.scattering_parameter == e5080b_no_device.settings.scattering_parameter

    def test_frequency_span_property(self, e5080b_no_device: E5080B):
        """Test the frequency span property"""
        assert hasattr(e5080b_no_device, "frequency_span")
        assert e5080b_no_device.frequency_span == e5080b_no_device.settings.frequency_span

    def test_frequency_center_property(self, e5080b_no_device: E5080B):
        """Test the frequency center property"""
        assert hasattr(e5080b_no_device, "frequency_center")
        assert e5080b_no_device.frequency_center == e5080b_no_device.settings.frequency_center

    def test_frequency_start_property(self, e5080b_no_device: E5080B):
        """Test the frequency start property"""
        assert hasattr(e5080b_no_device, "frequency_start")
        assert e5080b_no_device.frequency_start == e5080b_no_device.settings.frequency_start

    def test_frequency_stop_property(self, e5080b_no_device: E5080B):
        """Test the frequency stop property"""
        assert hasattr(e5080b_no_device, "frequency_stop")
        assert e5080b_no_device.frequency_stop == e5080b_no_device.settings.frequency_stop

    def test_averaging_enabled_property(self, e5080b_no_device: E5080B):
        """Test the averaging enabled property"""
        assert hasattr(e5080b_no_device, "averaging_enabled")
        assert e5080b_no_device.averaging_enabled == e5080b_no_device.settings.averaging_enabled

    def test_number_averages_propertyy(self, e5080b_no_device: E5080B):
        """Test the number of averages property"""
        assert hasattr(e5080b_no_device, "number_averages")
        assert e5080b_no_device.number_averages == e5080b_no_device.settings.number_averages

    def test_trigger_mode_property(self, e5080b_no_device: E5080B):
        """Test the trigger mode property"""
        assert hasattr(e5080b_no_device, "trigger_mode")
        assert e5080b_no_device.trigger_mode == e5080b_no_device.settings.trigger_mode

    def test_number_points_property(self, e5080b_no_device: E5080B):
        """Test the number points property"""
        assert hasattr(e5080b_no_device, "number_points")
        assert e5080b_no_device.number_points == e5080b_no_device.settings.number_points

    def test_electrical_delay_property(self, e5080b_no_device: E5080B):
        """Test the electrical delay property"""
        assert hasattr(e5080b_no_device, "electrical_delay")
        assert e5080b_no_device.electrical_delay == e5080b_no_device.settings.electrical_delay

    def test_device_timeout_property(self, e5080b_no_device: E5080B):
        """Test the device timeout property"""
        assert hasattr(e5080b_no_device, "device_timeout")
        assert e5080b_no_device.device_timeout == e5080b_no_device.settings.device_timeout


# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from typing import Any, Callable


class FluqeParameter:
    """Basic implementation of a qcodes like parameter class, but much simpler.
    Working, but still work in progress


    Args:
        name(str): name of the parameter
        value(any|none): initial value of the parameter
        set_method(callable|None): the set method of the parameter. If not is given, defaults to `set_raw()`.
    Calling
    """

    def __init__(self, name: str, value: Any | None = None, set_method: Callable | None = None):
        """init method"""
        # TODO: add units and other useful info
        self.name = name
        self._value = value
        self.set = set_method

    def __call__(self, *args: Any, **kwargs: Any) -> Any | None:
        """Call method for the parameter class"""
        # TODO: implement get_raw like in qcodes
        if not args and not kwargs:
            return self.get(*args, **kwargs)
        self._value = self.set(*args, **kwargs) if self.set is not None else self.set_raw(args[0])
        return None

    def get(self) -> Any:
        """Get method. Returns the parameter current value.

        Returns:
            Any: parameter value
        """
        return self._value

    def set_raw(self, value: Any) -> Any:
        """set_raw method. This is the default if no set method is given. The parameter is assigned the value given as argument.

        Args:
            value (Any): parameter value to set

        Returns:
            Any: parameter value set.
        """
        self._value = value
        return value



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Square waveform."""

import numpy as np

from qililab.qprogram.decorators import requires_domain
from qililab.qprogram.variable import Domain
from qililab.yaml import yaml

from .waveform import Waveform


@yaml.register_class
class Square(Waveform):
    """Square (rectangular) waveform. Given by a constant height line.

    Args:
        duration (int): Duration of the pulse (ns).
        amplitude (float): Maximum amplitude of the pulse.

    Examples:
        To get the envelope of a square waveform, with ``amplitude`` equal to ``X``, you need to do:

        .. code-block:: python

            import qililab as ql
            square_envelope = ql.Square( amplitude=X, duration=50).envelope()

        which for ``X`` being ``1.`` and ``0.75``, look respectively like:

        .. image:: /classes_images/rectangulars.png
            :width: 800
            :align: center
    """

    @requires_domain("amplitude", Domain.Voltage)
    @requires_domain("duration", Domain.Time)
    def __init__(self, amplitude: float, duration: int):
        super().__init__()
        self.amplitude = amplitude
        self.duration = duration

    def envelope(self, resolution: int = 1) -> np.ndarray:
        """Constant amplitude envelope.

        Args:
            resolution (float, optional): Resolution of the pulse. Defaults to 1.

        Returns:
            np.ndarray: Height of the envelope for each time step.
        """
        return self.amplitude * np.ones(round(self.duration / resolution))

    def get_duration(self) -> int:
        """Get the duration of the waveform.

        Returns:
            int: The duration of the waveform in ns.
        """
        return self.duration



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from qililab.qprogram.blocks.block import Block
from qililab.qprogram.variable import Variable
from qililab.yaml import yaml


@yaml.register_class
class ForLoop(Block):
    def __init__(self, variable: Variable, start: int | float, stop: int | float, step: int | float) -> None:
        super().__init__()
        self.variable: Variable = variable
        self.start: int | float = start
        self.stop: int | float = stop
        self.step: int | float = step



import numpy as np
import pytest

from qililab.waveforms import Chained, Ramp, Square


@pytest.fixture(name="chained_wf")
def fixture_square() -> Chained:
    return Chained(
        waveforms=[
            Ramp(from_amplitude=0.0, to_amplitude=1.0, duration=100),
            Square(amplitude=1.0, duration=200),
            Ramp(from_amplitude=1.0, to_amplitude=0.0, duration=100),
        ]
    )


class TestRamp:
    def test_init(self, chained_wf: Chained):
        """Test __init__ method"""
        assert isinstance(chained_wf.waveforms, list)
        assert len(chained_wf.waveforms) == 3

    def test_envelope_method(self, chained_wf: Chained):
        """Test envelope method"""
        expected_envelope = np.concatenate(
            [
                Ramp(from_amplitude=0.0, to_amplitude=1.0, duration=100).envelope(),
                Square(amplitude=1.0, duration=200).envelope(),
                Ramp(from_amplitude=1.0, to_amplitude=0.0, duration=100).envelope(),
            ]
        )
        envelope = chained_wf.envelope()
        assert np.allclose(envelope, expected_envelope)

    def test_get_duration_method(self, chained_wf: Chained):
        """Test get_duration method"""
        assert chained_wf.get_duration() == 400



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
from __future__ import annotations

from typing import TYPE_CHECKING, TypedDict

import numpy as np
from qpysequence import Acquisitions, Program, Waveforms
from qpysequence import Sequence as QpySequence
from qpysequence.constants import AWG_MAX_GAIN, INST_MAX_WAIT
from qpysequence.library import long_wait
from qpysequence.program import Block, Loop, Register
from qpysequence.program.instructions import (
    Acquire,
    Move,
    Play,
    ResetPh,
    SetAwgGain,
    SetMrk,
    SetPh,
    Stop,
    UpdParam,
    Wait,
)

from qililab.config import logger
from qililab.pulse.pulse_bus_schedule import PulseBusSchedule
from qililab.pulse.pulse_schedule import PulseSchedule
from qililab.typings import InstrumentName

if TYPE_CHECKING:
    from qililab.instruments.qblox import QbloxModule, QbloxSequencer
    from qililab.pulse.pulse_bus_schedule import PulseBusSchedule
    from qililab.pulse.pulse_schedule import PulseSchedule
    from qililab.pulse.pulse_shape.pulse_shape import PulseShape
    from qililab.settings.digital.digital_compilation_bus_settings import DigitalCompilationBusSettings


class ModuleSequencer(TypedDict):
    module: QbloxModule
    sequencer: QbloxSequencer


class QbloxCompiler:
    """Qblox compiler for pulse schedules. Its only public method is `compile`, which compiles a pulse schedule to qpysequences (see docs for `QBloxCompiler.compile`).
    The class object is meant to be initialized once, with `compile` running as many times as necessary. This way the class attributes do not have to be initialized
    at each single compilation.

    Args:
        platform (Platform): Platform object representing the laboratory setup used to control quantum devices.
    Raises:
        ValueError: at init if no readout module (QRM) is found in platform.
    """

    def __init__(
        self,
        buses: dict[str, DigitalCompilationBusSettings],
        module_and_sequencer_per_bus: dict[str, ModuleSequencer],
    ):
        self.buses = buses
        self.module_and_sequencer_per_bus = module_and_sequencer_per_bus
        self.nshots = 0
        self.num_bins = 0
        self.repetition_duration = 0
        self.readout_modules = [InstrumentName.QBLOX_QRM, InstrumentName.QRMRF]
        self.control_modules = [InstrumentName.QBLOX_QCM, InstrumentName.QCMRF]

    def compile(
        self, pulse_schedule: PulseSchedule, num_avg: int, repetition_duration: int, num_bins: int
    ) -> dict[str, list[QpySequence]]:
        """Compiles a given pulse schedule to a Sequence (QpySequence). If the compiler object is created and not
        reinitialized between executions, previously run Sequences and PulseBusSchedules are saved to the corresponding's
        qblox_module sequences and cache, respectively. They can then be loaded from the stored data instead of being
        compiled again.

        Args:
            pulse_schedule (PulseSchedule): PulseSchedule to be executed
            num_avg (int): hardware average - number of times to average execution over
            repetition_duration (int): fixed execution duration. Execution always lasts `repetition_duration` ns. If execution is shorter than that, a wait time of `repetition_duration` - circuit_execution_time is added (this is used for qubti relaxation)
            num_bins (int): number of shots to execute

        Raises:
            ValueError: if circuit execution time is longer than `repetition_duration`
            NotImplementedError: if more than one readout module is found connected to the platform

        Returns:
            dict[str, list[QpySequence]]: dictionary where the keys correspond to the bus which has to execute a list of QpySequences, given in the values
        """
        if num_avg != self.nshots or repetition_duration != self.repetition_duration or num_bins != self.num_bins:
            self.nshots = num_avg
            self.repetition_duration = repetition_duration
            self.num_bins = num_bins
            for bus_alias in self.module_and_sequencer_per_bus:
                self.module_and_sequencer_per_bus[bus_alias]["module"].clear_cache()

        bus_to_schedule = {schedule.bus_alias: schedule for schedule in pulse_schedule}

        compiled_sequences = {}  # type: dict[str, list[QpySequence]]

        # generally a sequencer_schedule is the schedule sent to a specific bus, except for readout,
        # where multiple schedules for different sequencers are sent to the same bus
        for bus_alias in bus_to_schedule:
            qblox_module = self.module_and_sequencer_per_bus[bus_alias]["module"]
            sequencer = self.module_and_sequencer_per_bus[bus_alias]["sequencer"]
            sequencer_schedule = bus_to_schedule[bus_alias]
            # check if circuit lasts longer than repetition duration
            end_time = None if len(sequencer_schedule.timeline) == 0 else sequencer_schedule.timeline[-1].end_time
            if end_time is not None and end_time > self.repetition_duration:
                raise ValueError(
                    f"Circuit execution time cannnot be longer than repetition duration but found circuit time {end_time } > {repetition_duration} for qubit {sequencer_schedule.qubit}"
                )
            # create empty list if key does not exist
            if bus_alias not in compiled_sequences:
                compiled_sequences[bus_alias] = []
            # if the schedule is already compiled get it from the module's sequences
            if sequencer_schedule == qblox_module.cache.get(
                sequencer.identifier
            ):  # if it's already cached then dont compile
                compiled_sequences[bus_alias].append(qblox_module.sequences[sequencer.identifier])
                # If the schedule is in the cache, delete the acquisition data (if uploaded) # FIXME: acquisitions should be deleted after acquisitions and not at compilation
                if qblox_module.name in self.readout_modules and hasattr(
                    qblox_module, "device"
                ):  # TODO: remove hasattr when the fixme above is done
                    qblox_module.device.delete_acquisition_data(sequencer=sequencer.identifier, all=True)

            else:
                # compile the sequences
                sequence = self._translate_pulse_bus_schedule(sequencer_schedule)
                compiled_sequences[bus_alias].append(sequence)
                qblox_module.cache[sequencer.identifier] = sequencer_schedule
                qblox_module.sequences[sequencer.identifier] = sequence

        return compiled_sequences

    def _translate_pulse_bus_schedule(self, pulse_bus_schedule: PulseBusSchedule) -> QpySequence:
        """Translate a pulse sequence into a Q1ASM program, a waveform dictionary and
        acquisitions dictionary (that is, a QpySequence sequence).

        Args:
            pulse_bus_schedule (PulseBusSchedule): Pulse bus schedule to translate.

        Returns:
            Sequence: Qblox Sequence object containing the program and waveforms.
        """
        waveforms = self._generate_waveforms(pulse_bus_schedule=pulse_bus_schedule)
        acquisitions = self._generate_acquisitions(pulse_bus_schedule=pulse_bus_schedule)
        program = self._generate_program(pulse_bus_schedule=pulse_bus_schedule, waveforms=waveforms)
        return QpySequence(program=program, waveforms=waveforms, acquisitions=acquisitions)

    def _generate_waveforms(self, pulse_bus_schedule: PulseBusSchedule):
        """Generate I and Q waveforms from a PulseSequence object.
        Args:
            pulse_bus_schedule (PulseBusSchedule): PulseSequence object.
        Returns:
            Waveforms: Waveforms object containing the generated waveforms.
        """
        waveforms = Waveforms()

        unique_pulses: list[tuple[int, PulseShape]] = []

        for pulse_event in pulse_bus_schedule.timeline:
            if (pulse_event.duration, pulse_event.pulse.pulse_shape) not in unique_pulses:
                unique_pulses.append((pulse_event.duration, pulse_event.pulse.pulse_shape))
                amp = pulse_event.pulse.amplitude
                sign = 1 if amp >= 0 else -1
                envelope = pulse_event.envelope(amplitude=sign * 1.0)
                real = np.real(envelope)
                imag = np.imag(envelope)
                pair = (real, imag)
                waveforms.add_pair(pair=pair, name=pulse_event.pulse.label())

        return waveforms

    def _generate_acquisitions(self, pulse_bus_schedule: PulseBusSchedule) -> Acquisitions:
        """Generate Acquisitions object, currently containing a single acquisition named "default", with num_bins = 1
        and index = 0.

        Args:
            sequencer (AWGQbloxSequencer): sequencer to which we generate the acquisitions dictionary
            timeline (list[PulseEvent]): time ordered list of pulse events to play

        Returns:
            Acquisitions: Acquisitions object.
        """
        acquisitions = Acquisitions()
        if self.buses[pulse_bus_schedule.bus_alias].is_readout():
            for i, pulse in enumerate(pulse_bus_schedule.timeline):
                acquisitions.add(name=f"acq_q{pulse.qubit}_{i}", num_bins=self.num_bins, index=i)
        return acquisitions

    def _generate_program(self, pulse_bus_schedule: PulseBusSchedule, waveforms: Waveforms) -> Program:
        """Generate Q1ASM program

        Args:
            pulse_sequence (PulseSequence): pulse sequence
            waveforms (Waveforms): waveforms
            sequencer (int): index of the sequencer to generate the program

        Returns:
            Program: Q1ASM program.
        """
        bus = self.buses[pulse_bus_schedule.bus_alias]
        qblox_module = self.module_and_sequencer_per_bus[pulse_bus_schedule.bus_alias]["module"]
        sequencer = self.module_and_sequencer_per_bus[pulse_bus_schedule.bus_alias]["sequencer"]
        MIN_WAIT = 4

        # Define program's blocks
        program = Program()
        start = Block(name="start")
        start.append_component(ResetPh())
        # Check if any markers should be set ON
        if qblox_module.name == InstrumentName.QCMRF:
            mask = int("".join(["1" if i in [0, 1] and i in sequencer.outputs else "0" for i in range(4)])[::-1], 2)
        elif qblox_module.name == InstrumentName.QRMRF:
            mask = int("".join(["1" if i in [1] and i - 1 in sequencer.outputs else "0" for i in range(4)])[::-1], 2)
        else:
            mask = 0
        start.append_component(SetMrk(mask))
        start.append_component(UpdParam(4))
        program.append_block(block=start)
        # Create registers with 0 and 1 (necessary for qblox)
        weight_registers = Register(), Register()
        if bus.is_readout():
            self._init_weights_registers(registers=weight_registers, program=program)
        avg_loop = Loop(name="average", begin=int(self.nshots))  # type: ignore
        bin_loop = Loop(name="bin", begin=0, end=self.num_bins, step=1)
        avg_loop.append_component(bin_loop)
        program.append_block(avg_loop)
        stop = Block(name="stop")
        # Set all markers to OFF
        stop.append_component(SetMrk(0))
        stop.append_component(UpdParam(4))
        stop.append_component(Stop())
        program.append_block(block=stop)
        timeline = pulse_bus_schedule.timeline
        if len(timeline) > 0 and timeline[0].start_time != 0:
            bin_loop.append_component(long_wait(wait_time=int(timeline[0].start_time)))

        for i, pulse_event in enumerate(timeline):
            waveform_pair = waveforms.find_pair_by_name(pulse_event.pulse.label())
            wait_time = timeline[i + 1].start_time - pulse_event.start_time if (i < (len(timeline) - 1)) else MIN_WAIT

            # Allow wait times longer than INST_MAX_WAIT
            long_wait_time = 0
            if wait_time > INST_MAX_WAIT:
                long_wait_time = wait_time - 2 * MIN_WAIT
                wait_time = 2 * MIN_WAIT
            # Add instrucitons to play the pulse
            phase = int((pulse_event.pulse.phase % (2 * np.pi)) * 1e9 / (2 * np.pi))
            gain = int(np.abs(pulse_event.pulse.amplitude) * AWG_MAX_GAIN)  # np.abs() needed for negative pulses
            bin_loop.append_component(SetAwgGain(gain_0=gain, gain_1=gain))
            bin_loop.append_component(SetPh(phase=phase))
            bin_loop.append_component(
                Play(
                    waveform_0=waveform_pair.waveform_i.index,
                    waveform_1=waveform_pair.waveform_q.index,
                    # wait until next pulse if QCM. If QRM wait min time (4) and wait time is added after acquiring
                    wait_time=int(wait_time) if not bus.is_readout() else MIN_WAIT,
                )
            )
            if bus.is_readout():
                self._append_acquire_instruction(
                    loop=bin_loop,
                    bin_index=bin_loop.counter_register,
                    acq_index=i,
                    wait=wait_time - MIN_WAIT if (i < len(timeline) - 1) else MIN_WAIT,
                )

            # Add long wait for wait time if necessary
            if long_wait_time != 0:
                bin_loop.append_component(long_wait(long_wait_time))

        if self.repetition_duration is not None:
            wait_time = self.repetition_duration - bin_loop.duration_iter
            if wait_time > MIN_WAIT:
                bin_loop.append_component(long_wait(wait_time=wait_time))

        logger.info("Q1ASM program: \n %s", repr(program))
        return program

    def _append_acquire_instruction(
        self,
        loop: Loop,
        bin_index: Register | int,
        acq_index: int,
        wait: int,
    ):
        """Append an acquire instruction to the loop."""
        # TODO: implement time of flight properly to remove this hack
        loop.append_component(Wait(wait_time=220))
        loop.append_component(Acquire(
                acq_index=acq_index,
                bin_index=bin_index,
                wait_time=wait,
            )
        )

    def _init_weights_registers(self, registers: tuple[Register, Register], program: Program):
        """Initialize the weights `registers` and place the required instructions in the setup block of the `program`."""
        move_0 = Move(0, registers[0])
        move_1 = Move(1, registers[1])
        setup_block = program.get_block(name="setup")
        setup_block.append_components([move_0, move_1], bot_position=1)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""QuantumMachinesResult class."""

from warnings import warn

import numpy as np

from qililab.result.qprogram.measurement_result import MeasurementResult
from qililab.typings.enums import ResultName
from qililab.yaml import yaml


@yaml.register_class
class QuantumMachinesMeasurementResult(MeasurementResult):
    """Contains the data obtained from a single measurment in Quantum Machines hardware.

    Args:
        I (np.ndarray): Data obtained from I stream.
        Q (np.ndarray): Data obtained from Q stream.
        adc1 (np.ndarray, optional): Data obtained from adc1 stream. Defaults to None.
        adc2 (np.ndarray, optional): Data obtained from adc2 stream. Defaults to None.
    """

    name = ResultName.QUANTUM_MACHINES_MEASUREMENT

    def __init__(
        self,
        bus: str,
        I: np.ndarray,
        Q: np.ndarray,
        adc1: np.ndarray | None = None,
        adc2: np.ndarray | None = None,
    ):
        super().__init__(bus=bus)
        self.I = I
        self.Q = Q
        self.adc1 = adc1
        self.adc2 = adc2
        self._classification_threshold = None

    def set_classification_threshold(self, classification_threshold):
        """Sets the `_classification_threshold` of the class."""
        self._classification_threshold = classification_threshold

    @property
    def array(self) -> np.ndarray:
        """Returns I/Q data as a combined numpy array.

        Returns:
            np.ndarray: The I/Q data as a compined nummpy array.
        """

        return np.concatenate((self.I.reshape(1, *self.I.shape), self.Q.reshape(1, *self.Q.shape)), axis=0)

    @property
    def threshold(self) -> np.ndarray:
        """Get the thresholded data as an np.ndarray. If the threshold is `None` thus not specified, returns an array of zeros.

        Returns:
            np.ndarray: The thresholded data.
        """
        if self._classification_threshold is None:
            warn("Classification threshold is not specified, returning a `np.zeros` array.", stacklevel=2)

        return (
            np.where(self._classification_threshold <= self.I, 1.0, 0.0)
            if self._classification_threshold is not None
            else np.zeros(self.I.shape)
        )



import copy
from unittest.mock import MagicMock, patch

import pytest

from qililab.instrument_controllers.qdevil.qdevil_qdac2_controller import QDevilQDac2Controller
from qililab.instruments.qdevil.qdevil_qdac2 import QDevilQDac2
from qililab.platform import Platform
from qililab.settings import Settings
from qililab.typings import ConnectionName
from tests.data import SauronQDevil
from tests.test_utils import build_platform


@pytest.fixture(name="platform")
def fixture_platform() -> Platform:
    """Return Platform object."""
    return build_platform(runcard=SauronQDevil.runcard)


@pytest.fixture(name="qdevil_qdac2_controller_wrong_module")
def fixture_qdevil_qdac2_controller_wrong_module_wrong_module(platform: Platform) -> QDevilQDac2Controller:
    """Return an instance of the QDAC-II controller with an incorrect module."""
    settings = copy.deepcopy(SauronQDevil.qdevil_qdac2_controller_wrong_module)
    settings.pop("name")
    return QDevilQDac2Controller(settings=settings, loaded_instruments=platform.instruments)


class TestQDevilQDac2Controller:
    """Unit tests checking the QDAC-II controller attributes and methods"""

    @pytest.mark.parametrize("controller_alias", ["qdac_controller_ip", "qdac_controller_usb"])
    def test_initialization(self, platform: Platform, controller_alias: str):
        """Test QDAC-II controller has been initialized correctly."""
        controller_instance = platform.instrument_controllers.get_instrument_controller(alias=controller_alias)
        assert isinstance(controller_instance, QDevilQDac2Controller)

        controller_settings = controller_instance.settings
        assert isinstance(controller_settings, Settings)

        controller_modules = controller_instance.modules
        assert len(controller_modules) == 1
        assert isinstance(controller_modules[0], QDevilQDac2)

    @patch("qililab.instrument_controllers.qdevil.qdevil_qdac2_controller.QDevilQDac2Device", autospec=True)
    @pytest.mark.parametrize("controller_alias", ["qdac_controller_ip", "qdac_controller_usb"])
    def test_initialize_device(self, device_mock: MagicMock, platform: Platform, controller_alias: str):
        """Test QDAC-II controller initializes device correctly."""
        controller_instance = platform.instrument_controllers.get_instrument_controller(alias=controller_alias)

        controller_instance._initialize_device()

        name = f"{controller_instance.name.value}"
        address = (
            f"TCPIP::{controller_instance.address}::5025::SOCKET"
            if controller_instance.connection.name == ConnectionName.TCP_IP
            else f"ASRL/dev/{controller_instance.address}::INSTR"
        )
        device_mock.assert_called_once_with(name, address)

    def test_check_supported_modules_raises_exception(
        self, qdevil_qdac2_controller_wrong_module: QDevilQDac2Controller
    ):
        """Test QDAC-II controller raises an error if initialized with wrong module."""
        with pytest.raises(ValueError):
            qdevil_qdac2_controller_wrong_module._check_supported_modules()



"""Tests for the Qblox Module class."""

from typing import cast
from unittest.mock import MagicMock

import pytest
from qblox_instruments.qcodes_drivers.module import Module as QcmQrm
from qblox_instruments.qcodes_drivers.sequencer import Sequencer
from qpysequence import Acquisitions, Program, Sequence, Waveforms, Weights

from qililab.data_management import build_platform
from qililab.instruments.instrument import ParameterNotFound
from qililab.instruments.qblox import QbloxQRMRF
from qililab.platform import Platform
from qililab.typings import Parameter

MAX_ATTENUATION = 30


@pytest.fixture(name="platform")
def fixture_platform():
    """platform fixture"""
    return build_platform(runcard="tests/instruments/qblox/qblox_runcard.yaml")


@pytest.fixture(name="qrm_rf")
def fixture_qrm(platform: Platform):
    qrm_rf = cast(QbloxQRMRF, platform.get_element(alias="qrm-rf"))

    sequencer_mock_spec = [
        *Sequencer._get_required_parent_attr_names(),
        "sync_en",
        "gain_awg_path0",
        "gain_awg_path1",
        "sequence",
        "mod_en_awg",
        "nco_freq",
        "scope_acq_sequencer_select",
        "channel_map_path0_out0_en",
        "channel_map_path1_out1_en",
        "demod_en_acq",
        "integration_length_acq",
        "mixer_corr_phase_offset_degree",
        "mixer_corr_gain_ratio",
        "connect_out0",
        "connect_out1",
        "connect_out2",
        "connect_out3",
        "marker_ovr_en",
        "offset_awg_path0",
        "offset_awg_path1",
        "connect_acq",
        "thresholded_acq_threshold",
        "thresholded_acq_rotation",
    ]

    module_mock_spec = [
        *QcmQrm._get_required_parent_qtm_attr_names(),
        "reference_source",
        "sequencer0",
        "sequencer1",
        "out0_offset",
        "out1_offset",
        "out2_offset",
        "out3_offset",
        "scope_acq_avg_mode_en_path0",
        "scope_acq_avg_mode_en_path1",
        "scope_acq_trigger_mode_path0",
        "scope_acq_trigger_mode_path1",
        "sequencers",
        "scope_acq_sequencer_select",
        "get_acquisitions",
        "disconnect_outputs",
        "disconnect_inputs",
        "arm_sequencer",
        "start_sequencer",
        "reset",
        "set",
        "out0_in0_lo_freq_cal_type_default",
    ]

    # Create a mock device using create_autospec to follow the interface of the expected device
    qrm_rf.device = MagicMock()
    qrm_rf.device.mock_add_spec(module_mock_spec)
    qrm_rf.device._get_max_out_att_0 = MagicMock(return_value=MAX_ATTENUATION)
    qrm_rf.device._run_mixer_lo_calib = MagicMock()

    qrm_rf.device.sequencers = {0: MagicMock(), 1: MagicMock()}

    for sequencer in qrm_rf.device.sequencers:
        qrm_rf.device.sequencers[sequencer].mock_add_spec(sequencer_mock_spec)
        qrm_rf.device.sequencers[sequencer].sideband_cal = MagicMock()

    return qrm_rf


class TestQbloxQRMRF:
    def test_init(self, qrm_rf: QbloxQRMRF):
        assert qrm_rf.alias == "qrm-rf"
        assert len(qrm_rf.awg_sequencers) == 2

    @pytest.mark.parametrize(
        "parameter, value",
        [
            # Test GAIN setting
            (Parameter.GAIN, 2.0),
            (Parameter.GAIN, 3.5),
            # Test GAIN_I and GAIN_Q settings
            (Parameter.GAIN_I, 1.5),
            (Parameter.GAIN_Q, 1.5),
            # Test OFFSET_I and OFFSET_Q settings
            (Parameter.OFFSET_I, 0.1),
            (Parameter.OFFSET_Q, 0.2),
            # Test IF setting (intermediate frequency)
            (Parameter.IF, 100e6),
            # Test HARDWARE_MODULATION setting
            (Parameter.HARDWARE_MODULATION, True),
            # Test GAIN_IMBALANCE setting
            (Parameter.GAIN_IMBALANCE, 0.05),
            # Test PHASE_IMBALANCE setting
            (Parameter.PHASE_IMBALANCE, 0.02),
            # QRM-RF specific
            (Parameter.LO_FREQUENCY, 5e9),
            (Parameter.OUT0_IN0_LO_FREQ, 5e9),
            (Parameter.OUT0_IN0_LO_EN, True),
            (Parameter.OUT0_ATT, 0.5),
            (Parameter.IN0_ATT, 0.5),
            (Parameter.OUT0_OFFSET_PATH0, 0.5),
            (Parameter.OUT0_OFFSET_PATH1, 6e9),
        ],
    )
    def test_set_parameter(self, qrm_rf: QbloxQRMRF, parameter, value):
        """Test setting parameters for QCM sequencers using parameterized values."""
        qrm_rf.set_parameter(parameter, value, channel_id=0)
        sequencer = qrm_rf.get_sequencer(0)

        # Check values based on the parameter
        if parameter == Parameter.GAIN:
            assert sequencer.gain_i == value
            assert sequencer.gain_q == value
        elif parameter == Parameter.GAIN_I:
            assert sequencer.gain_i == value
        elif parameter == Parameter.GAIN_Q:
            assert sequencer.gain_q == value
        elif parameter == Parameter.OFFSET_I:
            assert sequencer.offset_i == value
        elif parameter == Parameter.OFFSET_Q:
            assert sequencer.offset_q == value
        elif parameter == Parameter.IF:
            assert sequencer.intermediate_frequency == value
        elif parameter == Parameter.HARDWARE_MODULATION:
            assert sequencer.hardware_modulation == value
        elif parameter == Parameter.GAIN_IMBALANCE:
            assert sequencer.gain_imbalance == value
        elif parameter == Parameter.PHASE_IMBALANCE:
            assert sequencer.phase_imbalance == value
        elif parameter == Parameter.OUT0_IN0_LO_FREQ:
            assert qrm_rf.settings.out0_in0_lo_freq == value
        elif parameter == Parameter.OUT0_IN0_LO_EN:
            assert qrm_rf.settings.out0_in0_lo_en == value
        elif parameter == Parameter.OUT0_ATT:
            assert qrm_rf.settings.out0_att == value
        elif parameter == Parameter.IN0_ATT:
            assert qrm_rf.settings.in0_att == value
        elif parameter == Parameter.OUT0_OFFSET_PATH0:
            assert qrm_rf.settings.out0_offset_path0 == value
        elif parameter == Parameter.OUT0_OFFSET_PATH0:
            assert qrm_rf.settings.out0_offset_path1 == value

    @pytest.mark.parametrize(
        "parameter, value",
        [
            # Invalid parameter (should raise ParameterNotFound)
            (Parameter.BUS_FREQUENCY, 42),  # Invalid parameter
        ],
    )
    def test_set_parameter_raises_error(self, qrm_rf: QbloxQRMRF, parameter, value):
        """Test setting parameters for QCM sequencers using parameterized values."""
        with pytest.raises(ParameterNotFound):
            qrm_rf.set_parameter(parameter, value, channel_id=0)

        with pytest.raises(Exception):
            qrm_rf.set_parameter(Parameter.OUT0_ATT, value=MAX_ATTENUATION + 10, channel_id=None)

    @pytest.mark.parametrize(
        "parameter, expected_value",
        [
            # Test GAIN_I and GAIN_Q settings
            (Parameter.GAIN_I, 1.0),
            (Parameter.GAIN_Q, 1.0),
            # Test OFFSET_I and OFFSET_Q settings
            (Parameter.OFFSET_I, 0.0),
            (Parameter.OFFSET_Q, 0.0),
            # Test IF setting (intermediate frequency)
            (Parameter.IF, 100e6),
            # Test HARDWARE_MODULATION setting
            (Parameter.HARDWARE_MODULATION, True),
            # Test GAIN_IMBALANCE setting
            (Parameter.GAIN_IMBALANCE, 0.05),
            # Test PHASE_IMBALANCE setting
            (Parameter.PHASE_IMBALANCE, 0.02),
            # QRM-RF specific
            (Parameter.LO_FREQUENCY, 3e9),
            (Parameter.OUT0_IN0_LO_FREQ, 3e9),
            (Parameter.OUT0_IN0_LO_EN, True),
            (Parameter.OUT0_ATT, 10),
            (Parameter.IN0_ATT, 2),
            (Parameter.OUT0_OFFSET_PATH0, 0.2),
            (Parameter.OUT0_OFFSET_PATH1, 0.07),
        ],
    )
    def test_get_parameter(self, qrm_rf: QbloxQRMRF, parameter, expected_value):
        """Test setting parameters for QCM sequencers using parameterized values."""
        value = qrm_rf.get_parameter(parameter, channel_id=0)
        assert value == expected_value

    def test_get_parameter_raises_error(self, qrm_rf: QbloxQRMRF):
        """Test setting parameters for QCM sequencers using parameterized values."""
        with pytest.raises(ParameterNotFound):
            qrm_rf.get_parameter(Parameter.BUS_FREQUENCY, channel_id=0)

    @pytest.mark.parametrize(
        "channel_id, expected_error",
        [
            (0, None),  # Valid channel ID
            (5, Exception),  # Invalid channel ID
        ],
    )
    def test_invalid_channel(self, qrm_rf: QbloxQRMRF, channel_id, expected_error):
        """Test handling invalid channel IDs when setting parameters."""
        if expected_error:
            with pytest.raises(expected_error):
                qrm_rf.set_parameter(Parameter.GAIN, 2.0, channel_id=channel_id)
        else:
            qrm_rf.set_parameter(Parameter.GAIN, 2.0, channel_id=channel_id)
            sequencer = qrm_rf.get_sequencer(channel_id)
            assert sequencer.gain_i == 2.0
            assert sequencer.gain_q == 2.0

    def test_initial_setup(self, qrm_rf: QbloxQRMRF):
        """Test the initial setup of the QCM module."""
        qrm_rf.initial_setup()

        # Verify the correct setup calls were made on the device
        qrm_rf.device.disconnect_outputs.assert_called_once()
        qrm_rf.device.disconnect_inputs.assert_called_once()
        for sequencer in qrm_rf.awg_sequencers:
            qrm_rf.device.sequencers[sequencer.identifier].connect_out0.assert_called_with("IQ")
            qrm_rf.device.sequencers[sequencer.identifier].connect_acq.assert_called_with("in0")
            qrm_rf.device.sequencers[sequencer.identifier].sync_en.assert_called_with(False)

    def test_run(self, qrm_rf: QbloxQRMRF):
        """Test running the QCM module."""
        qrm_rf.sequences[0] = Sequence(
            program=Program(), waveforms=Waveforms(), acquisitions=Acquisitions(), weights=Weights()
        )
        qrm_rf.run(channel_id=0)

        sequencer = qrm_rf.get_sequencer(0)
        qrm_rf.device.arm_sequencer.assert_called_with(sequencer=sequencer.identifier)
        qrm_rf.device.start_sequencer.assert_called_with(sequencer=sequencer.identifier)

    def test_upload_qpysequence(self, qrm_rf: QbloxQRMRF):
        """Test uploading a QpySequence to the QCM module."""
        sequence = Sequence(program=Program(), waveforms=Waveforms(), acquisitions=Acquisitions(), weights=Weights())
        qrm_rf.upload_qpysequence(qpysequence=sequence, channel_id=0)

        qrm_rf.device.sequencers[0].sequence.assert_called_once_with(sequence.todict())

    def test_clear_cache(self, qrm_rf: QbloxQRMRF):
        """Test clearing the cache of the QCM module."""
        qrm_rf.cache = {0: MagicMock()}  # type: ignore[misc]
        qrm_rf.clear_cache()

        assert qrm_rf.cache == {}
        assert qrm_rf.sequences == {}

    def test_reset(self, qrm_rf: QbloxQRMRF):
        """Test resetting the QCM module."""
        qrm_rf.reset()

        qrm_rf.device.reset.assert_called_once()
        assert qrm_rf.cache == {}
        assert qrm_rf.sequences == {}

    def test_calibrate_mixers(self, qrm_rf: QbloxQRMRF):
        """Test calibrating the QRM mixers."""
        channel_id = 0
        cal_type = "lo"

        qrm_rf.calibrate_mixers(cal_type=cal_type, channel_id=channel_id)
        qrm_rf.device._run_mixer_lo_calib.assert_called_with(channel_id)

        cal_type = "lo_and_sidebands"

        qrm_rf.calibrate_mixers(cal_type=cal_type, channel_id=channel_id)
        qrm_rf.device._run_mixer_lo_calib.assert_called_with(channel_id)

        for sequencer in qrm_rf.device.sequencers:
            qrm_rf.device.sequencers[sequencer].sideband_cal.assert_called()

        cal_type = "oh hi Mark!"

        with pytest.raises(Exception):
            qrm_rf.calibrate_mixers(cal_type=cal_type, channel_id=channel_id)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""SignalGenerator class."""
from dataclasses import dataclass

from qililab.instruments.instrument import Instrument


class SignalGenerator(Instrument):
    """Abstract base class defining all instruments used to generate signals."""

    @dataclass
    class SignalGeneratorSettings(Instrument.InstrumentSettings):
        """Contains the settings of a specific signal generator.

        Args:
            power (float): Power of the instrument. Value range is (-120, 25).
            frequency (float): Frequency of the instrument. Value range is (1e6, 20e9).
        """

        power: float
        frequency: float
        rf_on: bool

    settings: SignalGeneratorSettings

    @property
    def power(self):
        """SignalGenerator 'power' property.

        Returns:
            float: settings.power.
        """
        return self.settings.power

    @property
    def frequency(self):
        """SignalGenerator 'frequency' property.

        Returns:
            float: settings.frequency.
        """
        return self.settings.frequency

    @property
    def rf_on(self):
        """SignalGenerator 'rf_on' property.
        Returns:
            bool: settings.rf_on.
        """
        return self.settings.rf_on

    def to_dict(self):
        """Return a dict representation of the SignalGenerator class."""
        return dict(super().to_dict().items())



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""KeySight E5080B Instrument Controller"""

from dataclasses import dataclass
from typing import Sequence

from qililab.instrument_controllers.utils.instrument_controller_factory import InstrumentControllerFactory
from qililab.instrument_controllers.vector_network_analyzer.vector_network_analyzer_controller import (
    VectorNetworkAnalyzerController,
)
from qililab.instruments.keysight.e5080b_vna import E5080B
from qililab.typings.enums import InstrumentControllerName, InstrumentName
from qililab.typings.instruments.vector_network_analyzer import VectorNetworkAnalyzerDriver


@InstrumentControllerFactory.register
class E5080BController(VectorNetworkAnalyzerController):
    """KeySight E5080B Instrument Controller

    Args:
        name (InstrumentControllerName): Name of the Instrument Controller.
        device (RohdeSchwarz_E5080B): Instance of the qcodes E5080B class.
        settings (E5080BSettings): Settings of the instrument.
    """

    name = InstrumentControllerName.KEYSIGHT_E5080B
    device: VectorNetworkAnalyzerDriver
    modules: Sequence[E5080B]

    @dataclass
    class E5080BControllerSettings(VectorNetworkAnalyzerController.VectorNetworkAnalyzerControllerSettings):
        """Contains the settings of a specific E5080B Controller."""

    settings: E5080BControllerSettings

    def _initialize_device(self):
        """Initialize device attribute to the corresponding device class."""
        self.device = VectorNetworkAnalyzerDriver(
            name=f"{self.name.value}_{self.alias}", address=self.address, timeout=self.timeout
        )

    def _check_supported_modules(self):
        """check if all instrument modules loaded are supported modules for the controller."""
        for module in self.modules:
            if not isinstance(module, E5080B):
                raise ValueError(
                    f"Instrument {type(module)} not supported."
                    + f"The only supported instrument is {InstrumentName.KEYSIGHT_E5080B}"
                )



"""Tests for the PulseSequence class."""
import numpy as np
import pytest

from qililab.pulse import Gaussian, Pulse, PulseBusSchedule, PulseEvent


@pytest.fixture(name="pulse_event")
def fixture_pulse_event() -> PulseEvent:
    """Load PulseEvent.

    Returns:
        PulseEvent: Instance of the PulseEvent class.
    """
    pulse_shape = Gaussian(num_sigmas=4)
    pulse = Pulse(amplitude=1, phase=0, duration=50, frequency=1e9, pulse_shape=pulse_shape)
    return PulseEvent(pulse=pulse, start_time=0)


@pytest.fixture(name="pulse_bus_schedule")
def fixture_pulse_bus_schedule(pulse_event: PulseEvent) -> PulseBusSchedule:
    """Return PulseBusSchedule instance."""
    return PulseBusSchedule(bus_alias="readout_bus", timeline=[pulse_event])


@pytest.fixture(name="mux_pulse_bus_schedule")
def fixture_mux_pulse_bus_schedule() -> PulseBusSchedule:
    """Return multiplexed PulseBusSchedule instance."""
    pulse_event_1 = PulseEvent(
        pulse=Pulse(amplitude=1, phase=0.0, duration=1000, frequency=7.0, pulse_shape=Gaussian(num_sigmas=5)),
        start_time=0,
    )
    pulse_event_2 = PulseEvent(
        pulse=Pulse(amplitude=1, phase=0.0, duration=1000, frequency=7.1, pulse_shape=Gaussian(num_sigmas=5)),
        start_time=0,
    )
    return PulseBusSchedule(timeline=[pulse_event_1, pulse_event_2], port=0)


class TestPulseBusSchedule:
    """Unit tests checking the PulseSequence attributes and methods"""

    def test_add_event_method(self, pulse_bus_schedule: PulseBusSchedule, pulse_event: PulseEvent):
        """Test add method."""
        pulse_bus_schedule.add_event(pulse_event=pulse_event)

    def test_waveforms_method(self, pulse_bus_schedule: PulseBusSchedule):
        """Test waveforms method."""
        waveforms = pulse_bus_schedule.waveforms(resolution=0.1)
        assert isinstance(waveforms.i, np.ndarray) and isinstance(waveforms.q, np.ndarray)

    def test_iter_method(self, pulse_bus_schedule: PulseBusSchedule):
        """Test __iter__ method."""
        for pulse in pulse_bus_schedule:
            assert isinstance(pulse, PulseEvent)

    def test_unique_pulses_duration(self, pulse_bus_schedule: PulseBusSchedule):
        """Test the unique_pulses_duration property."""
        pulses_duration = sum(pulse.duration for pulse in pulse_bus_schedule.pulses)
        assert pulses_duration == pulse_bus_schedule.unique_pulses_duration

    def test_end(self, pulse_bus_schedule: PulseBusSchedule):
        """Test the end_time property."""
        last_pulse_event = pulse_bus_schedule.timeline[-1]
        end = last_pulse_event.start_time + last_pulse_event.pulse.duration
        assert end == pulse_bus_schedule.end_time

    def test_start(self, pulse_bus_schedule: PulseBusSchedule):
        """Test the start_time property."""
        first_pulse_event = pulse_bus_schedule.timeline[0]
        start = first_pulse_event.start_time
        assert start == pulse_bus_schedule.start_time

    def test_total_duration(self, pulse_bus_schedule: PulseBusSchedule):
        """Test the total duration property."""
        duration = pulse_bus_schedule.end_time - pulse_bus_schedule.start_time
        assert pulse_bus_schedule.duration == duration



from datetime import datetime
from unittest.mock import MagicMock, call, patch

import networkx as nx
import pandas as pd
import pytest

from qililab import Parameter
from qililab.calibration import CalibrationController, CalibrationNode
from qililab.data_management import build_platform
from qililab.platform.platform import Platform

# flake8: noqa


#################################################################################
#################################### SET UPS ####################################
#################################################################################

path_runcard = "tests/calibration/galadriel.yml"


def dummy_comparison_model(obtained: dict, comparison: dict) -> float:
    """Basic comparison model for testing."""
    return abs(sum(obtained["y"]) - sum(comparison["y"]))


######################
### NODES CREATION ###
######################
zeroth = CalibrationNode(
    nb_path="tests/calibration/notebook_test/zeroth.ipynb",
    qubit_index=[0, 1],  # qubit_index as list
)
first = CalibrationNode(
    nb_path="tests/calibration/notebook_test/first.ipynb",
    qubit_index=0,
)
second = CalibrationNode(
    nb_path="tests/calibration/notebook_test/second.ipynb",
    qubit_index=0,
)
third = CalibrationNode(
    nb_path="tests/calibration/notebook_test/third.ipynb",
    qubit_index=0,
)
fourth = CalibrationNode(
    nb_path="tests/calibration/notebook_test/fourth.ipynb",
    # no qubit index
)

# NODE MAPPING TO THE GRAPH (key = name in graph, value = node object):
nodes = {"zeroth_q0q1": zeroth, "first_q0": first, "second_q0": second, "third_q0": third, "fourth": fourth}


#######################
### GRAPHS CREATION ###
#######################

# fmt: off
# GOOD GRAPH CREATION:
G0 = nx.DiGraph()                           #        3 -->\
G0.add_edge("third_q0", "fourth")           #              \
G0.add_edge("second_q0", "fourth")          # 0 ---> 2 ---> 4
G0.add_edge("zeroth_q0q1", "second_q0")     #  \
G0.add_edge( "zeroth_q0q1", "first_q0")     #   \--> 1

# GOOD GRAPH CREATION:
G1 = nx.DiGraph()                           #        3 -->\
G1.add_edge("third_q0", "fourth")           #        ^     \
G1.add_edge("second_q0", "fourth")          # 0 ---> 2 ---> 4
G1.add_edge("zeroth_q0q1", "second_q0")     #  \     ^
G1.add_edge("zeroth_q0q1", "first_q0")      #   \--> 1
G1.add_edge("second_q0", "third_q0")
G1.add_edge("first_q0", "second_q0")

# GOOD GRAPH CREATION:
G2 = nx.DiGraph()                           #   /--> 3 ---> 4
G2.add_edge("third_q0", "fourth")           #  /
G2.add_edge("zeroth_q0q1", "third_q0")      # 0 ---> 2
G2.add_edge("zeroth_q0q1", "second_q0")     #  \
G2.add_edge("zeroth_q0q1", "first_q0")      #   \--> 1

# GOOD GRAPH CREATION:
G3 = nx.DiGraph()
G3.add_edge("third_q0", "fourth")           #   /--> 3 ---> 4
G3.add_edge("zeroth_q0q1", "third_q0")      #  /     ^
G3.add_edge("second_q0", "third_q0")        # 0 ---> 2
G3.add_edge("zeroth_q0q1", "second_q0")     #  \
G3.add_edge("zeroth_q0q1", "first_q0")      #   \--> 1

# GOOD GRAPH CREATION:
G4 = nx.DiGraph()
G4.add_edge("third_q0", "fourth")           #   /--> 2 -->\
G4.add_edge("second_q0", "third_q0")        #  /           \
G4.add_edge("first_q0", "third_q0")         # 0             3 ---> 4
G4.add_edge("zeroth_q0q1", "second_q0")     #  \           /
G4.add_edge("zeroth_q0q1", "first_q0")      #   \--> 1 -->/

# GOOD GRAPH CREATION:
G5 = nx.DiGraph()
G5.add_edge("third_q0", "fourth")           #   /--> 2 -->\
G5.add_edge("second_q0", "third_q0")        #  /     ^     \
G5.add_edge("first_q0", "third_q0")         # 0      |      3 ---> 4
G5.add_edge("zeroth_q0q1", "second_q0")     #  \     |     /
G5.add_edge("first_q0", "second_q0")        #   \--> 1 -->/
G5.add_edge("zeroth_q0q1", "first_q0")

# GOOD GRAPH CREATION:
G6 = nx.DiGraph()
G6.add_edge("third_q0", "fourth")           #   /--> 3 -->\
G6.add_edge("second_q0", "third_q0")        #  /     ^     \
G6.add_edge("zeroth_q0q1", "third_q0")      # 0 ---> 2      4
G6.add_edge("zeroth_q0q1", "second_q0")     #  \     v     /
G6.add_edge("second_q0", "first_q0")        #   \--> 1 -->/
G6.add_edge("zeroth_q0q1", "first_q0")
G6.add_edge("first_q0", "fourth")

# GOOD GRAPH CREATION:
G7 = nx.DiGraph()
G7.add_edge("third_q0", "fourth")           #   /--> 3 -->\
G7.add_edge("second_q0", "third_q0")        #  /     ^     \
G7.add_edge("zeroth_q0q1", "third_q0")      # 0 ---> 2      4
G7.add_edge("zeroth_q0q1", "second_q0")     #  \     ^     /
G7.add_edge("first_q0", "second_q0")        #   \--> 1 -->/
G7.add_edge("zeroth_q0q1", "first_q0")
G7.add_edge("first_q0", "fourth")

# GOOD GRAPH CREATION:
G8 = nx.DiGraph()
G8.add_edge("third_q0", "fourth")           #   /--> 3 -->\
G8.add_edge("second_q0", "third_q0")        #  /     ^     \
G8.add_edge("zeroth_q0q1", "third_q0")      # 0 ---> 2 ---> 4
G8.add_edge("zeroth_q0q1", "second_q0")     #  \     v     /
G8.add_edge("second_q0", "first_q0")        #   \--> 1 -->/
G8.add_edge("zeroth_q0q1", "first_q0")
G8.add_edge("second_q0", "fourth")
G8.add_edge("first_q0", "fourth")

# GOOD GRAPH CREATION:                      #        |
G9 = nx.DiGraph()                           #        v
G9.add_edge("third_q0", "fourth")           #   /--> 3 -->\
G9.add_edge("second_q0", "third_q0")        #  /     ^     \
G9.add_edge("zeroth_q0q1", "third_q0")      # 0 ---> 2 ---> 4
G9.add_edge("zeroth_q0q1", "second_q0")     #  \     v     /
G9.add_edge("second_q0", "first_q0")        #   \--> 1 -->/
G9.add_edge("zeroth_q0q1", "first_q0")      #        |
G9.add_edge("second_q0", "fourth")          #        v
G9.add_edge("first_q0", "fourth")
G9.add_edge("first_q0", "third_q0")

# BAD GRAPH CREATION:
B = nx.DiGraph()                            #         /---<---\
B.add_edge("third_q0", "fourth")            #        /         \
B.add_edge("second_q0", "third_q0")         # 0 --> 1 --> 2 --> 3 --> 4
B.add_edge("first_q0", "second_q0")
B.add_edge("zeroth_q0q1", "first_q0")
B.add_edge("third_q0", "first_q0")

good_graphs = [G0, G1, G2, G3, G4, G5, G6, G7, G8, G9]

# Leaves to Roots recursive calls of each graph. Examples: Calibrate and update_param.
G0_calls = [call(third), call(zeroth), call(second), call(fourth)]
G1_calls = [call(zeroth), call(first), call(second), call(third), call(fourth)]
G2_calls = [call(zeroth), call(third), call(fourth)]
G3_calls = [call(zeroth), call(second), call(third), call(fourth)]
G4_calls = [call(zeroth), call(second), call(first), call(third), call(fourth)]
G5_calls = [call(zeroth), call(first), call(second), call(third), call(fourth)]
G6_calls = [call(zeroth), call(second), call(third), call(first), call(fourth)]
G7_calls = [call(zeroth), call(first), call(second), call(third), call(fourth)]
G8_calls = [call(zeroth), call(second), call(third), call(first), call(fourth)]
G9_calls = [call(zeroth), call(second), call(first), call(third), call(fourth)]

leaves_to_roots_good_graphs_calls = [G0_calls,G1_calls ,G2_calls,G3_calls,G4_calls,G5_calls,G6_calls,G7_calls,G8_calls,G9_calls]
# fmt: on


##########################
### MOCKED CONTROLLERS ###
##########################
class RunAutomaticCalibrationMockedController(CalibrationController):
    """``CalibrationController`` to test the workflow of ``run_automatic_calibration()``, where its mocked ``calibrate_all()``, ``get_last_set_parameters()`` and ``get_last_fidelities()``."""

    def __init__(self, node_sequence, calibration_graph, runcard):
        super().__init__(node_sequence=node_sequence, calibration_graph=calibration_graph, runcard=runcard)
        self.calibrate_all = MagicMock(return_value=None)
        self.get_qubits_tables = MagicMock(return_value=(10, 10))
        self.get_last_set_parameters = MagicMock(
            return_value={("test", "test"): (0.0, "test", datetime.fromtimestamp(1999))}
        )
        self.get_last_fidelities = MagicMock(return_value={"test": (0.0, "test", datetime.fromtimestamp(1999))})


class CalibrateAllMockedController(CalibrationController):
    """``CalibrationController`` to test the workflow of ``calibrate_all()`` where its mocked ``calibrate()`` and ``update_parameters()``."""

    def __init__(self, node_sequence, calibration_graph, runcard):
        super().__init__(node_sequence=node_sequence, calibration_graph=calibration_graph, runcard=runcard)
        self.calibrate = MagicMock(return_value=None)
        self._update_parameters = MagicMock(return_value=None)


#################################################################################
############################## TESTS FOR THE CLASS ##############################
#################################################################################


###########################
### TEST INITIALIZATION ###
###########################
class TestInitializationCalibrationController:
    """Unit tests for the CalibrationController class initialization."""

    @pytest.mark.parametrize(
        "controller",
        [
            (graph, CalibrationController(node_sequence=nodes, calibration_graph=graph, runcard=path_runcard))
            for graph in good_graphs
        ],
    )
    def test_good_init_method(self, controller):
        """Test a valid initialization of the class."""
        # Assert:
        assert controller[1].calibration_graph == controller[0]
        assert isinstance(controller[1].calibration_graph, nx.DiGraph)
        assert controller[1].node_sequence == nodes
        assert isinstance(controller[1].node_sequence, dict)
        assert controller[1].runcard == path_runcard
        assert isinstance(controller[1].runcard, str)
        assert controller[1].platform.to_dict() == build_platform(path_runcard).to_dict()
        assert isinstance(controller[1].platform, Platform)
        assert controller[1].drift_timeout == 7200

    def test_bad_init_method(self):
        """Test an invalid initialization of the class.

        This happens when the graph is not a Direct Acyclic Graph.
        """
        # Assert:
        with pytest.raises(ValueError) as error:
            _ = CalibrationController(node_sequence=nodes, calibration_graph=B, runcard=path_runcard)

        assert str(error.value) == "The calibration graph must be a Directed Acyclic Graph (DAG)."


######################################
### TEST RUN AUTOMATIC CALIBRATION ###
######################################
@pytest.mark.parametrize(
    "controller",
    [
        RunAutomaticCalibrationMockedController(node_sequence=nodes, calibration_graph=graph, runcard=path_runcard)
        for graph in good_graphs
    ],
)
class TestRunAutomaticCalibrationFromCalibrationController:
    """Test that ``run_autoamtic_calibration()`` of ``CalibrationController`` behaves well."""

    def test_run_automatic_calibration(self, controller):
        """Test that `run_automatic_calibration()` gets the proper nodes to calibrate_all()."""
        # Act:
        output_dict = controller.run_automatic_calibration()

        # Asserts:
        controller.get_last_set_parameters.assert_called_once_with()
        controller.get_last_fidelities.assert_called_once_with()
        controller.get_qubits_tables.assert_called_once_with()

        assert output_dict == {
            "1q_table": 10,
            "2q_table": 10,
            "set_parameters": {("test", "test"): (0.0, "test", datetime.fromtimestamp(1999))},
            "fidelities": {"test": (0.0, "test", datetime.fromtimestamp(1999))},
        }

        # sourcery skip: extract-duplicate-method
        if controller.calibration_graph in [G0, G3]:
            controller.calibrate_all.assert_any_call(fourth)
            controller.calibrate_all.assert_any_call(first)
            assert controller.calibrate_all.call_count == 2

        elif controller.calibration_graph == G2:
            controller.calibrate_all.assert_any_call(fourth)
            controller.calibrate_all.assert_any_call(second)
            controller.calibrate_all.assert_any_call(first)
            assert controller.calibrate_all.call_count == 3

        elif controller.calibration_graph in [G1, G4, G5, G6, G7, G8, G9]:
            controller.calibrate_all.assert_any_call(fourth)
            assert controller.calibrate_all.call_count == 1


##########################
### TEST CALIBRATE ALL ###
##########################
@pytest.mark.parametrize(
    "controller",
    [
        (
            graph,
            expected_call_order,
            CalibrateAllMockedController(node_sequence=nodes, calibration_graph=graph, runcard=path_runcard),
        )
        for graph, expected_call_order in zip(good_graphs, leaves_to_roots_good_graphs_calls)
    ],
)
class TestCalibrateAllFromCalibrationController:
    """Test that ``calibrate_all()`` of ``CalibrationConroller`` behaves well."""

    def test_low_level_mockings_working_properly(self, controller):
        """Test that the mockings are working properly."""
        # Assert:
        assert all(node.previous_timestamp is None for node in controller[2].node_sequence.values())
        assert controller[2].calibrate() is None
        assert controller[2]._update_parameters() is None

    def test_calls_for_linear_calibration(self, controller):
        """Test that ``calibrate_all`` follows the correct logic for each graph, from leaves up to the roots"""

        # Reset mock calls:
        controller[2].calibrate.reset_mock()
        controller[2]._update_parameters.reset_mock()
        for node in controller[2].node_sequence.values():
            node.been_calibrated = False

        # Act:
        controller[2].calibrate_all(fourth)

        # Assert that 0, 3 & 4 notebooks have been calibrated:
        # (1 and 2 are calibrated in some graphs and not in others)
        for node in [zeroth, third, fourth]:
            assert node.been_calibrated is True

        # Asserts recursive calls
        controller[2].calibrate.assert_has_calls(controller[1])
        controller[2]._update_parameters.assert_has_calls(controller[1])


@pytest.mark.parametrize(
    "controller",
    [
        CalibrationController(node_sequence=nodes, calibration_graph=graph, runcard=path_runcard)
        for graph in good_graphs
    ],
)
class TestCalibrationController:
    """Test that the rest of ``CalibrationController`` methods behave well."""

    ######################
    ### TEST CALIBRATE ###
    ######################
    @patch("qililab.calibration.calibration_node.CalibrationNode.run_node")
    @patch("qililab.calibration.calibration_node.CalibrationNode._add_string_to_checked_nb_name")
    def test_calibrate(self, mock_add_str, mock_run, controller):
        """Test that the calibration method, calls node.run_node()."""
        for node in controller.node_sequence.values():
            controller.calibrate(node)
        assert mock_run.call_count == len(controller.node_sequence)
        assert mock_add_str.call_count == len(controller.node_sequence)

    ##############################
    ### TEST UPDATE PARAMETERS ###
    ##############################

    @patch("qililab.platform.Platform.set_parameter")
    @patch("qililab.calibration.calibration_controller.save_platform")
    def test_update_parameters(self, mock_save_platform, mock_set_params, controller):
        """Test that the update parameters method, calls ``platform.set_parameter()`` and ``save_platform()``."""
        for node in controller.node_sequence.values():
            node.output_parameters = {
                "platform_parameters": [
                    (Parameter.AMPLITUDE, 0, "test_bus", node.qubit_index),
                    (Parameter.AMPLITUDE, 1, "test_bus2", node.qubit_index),
                ]
            }
            controller._update_parameters(node)

            mock_set_params.assert_called_with(
                parameter=Parameter.AMPLITUDE, value=1, alias="test_bus2", channel_id=node.qubit_index
            )  # Checking the last call of the 2 there are.
            mock_save_platform.assert_called_with(controller.runcard, controller.platform)  # Checking the save call

        assert mock_set_params.call_count == 2 * len(controller.node_sequence)
        assert mock_save_platform.call_count == len(controller.node_sequence)

    ####################################
    ### TEST GET LAST SET PARAMETERS ###
    ####################################
    def test_get_last_set_parameters(self, controller):
        """Test that the ``get_last_set_parameters()`` method, gets the correct parameters."""
        for i, node in controller.node_sequence.items():
            node.output_parameters = {
                "platform_parameters": [("param", 0, f"test_bus_{i}", 0), ("param", 1, f"test_bus_{i}", 1)],
                "fidelities": [(0, f"param_{i}", 1), (1, f"param_{i}", 0.967)],
            }
            node.previous_timestamp = 1999

        df = controller.get_last_set_parameters()

        # Create the pandas DataFrame to test
        data = [
            [0, "zeroth_q0q1", datetime.fromtimestamp(1999)],
            [1, "zeroth_q0q1", datetime.fromtimestamp(1999)],
            [0, "first_q0", datetime.fromtimestamp(1999)],
            [1, "first_q0", datetime.fromtimestamp(1999)],
            [0, "second_q0", datetime.fromtimestamp(1999)],
            [1, "second_q0", datetime.fromtimestamp(1999)],
            [0, "third_q0", datetime.fromtimestamp(1999)],
            [1, "third_q0", datetime.fromtimestamp(1999)],
            [0, "fourth", datetime.fromtimestamp(1999)],
            [1, "fourth", datetime.fromtimestamp(1999)],
        ]
        idx = pd.MultiIndex.from_product(
            [
                ["param"],
                [
                    "test_bus_zeroth_q0q1",
                    "test_bus_first_q0",
                    "test_bus_second_q0",
                    "test_bus_third_q0",
                    "test_bus_fourth",
                ],
                [0, 1],
            ],
            names=["parameter", "bus", "qubit"],
        )
        col = ["value", "node_id", "datetime"]
        test_df = pd.DataFrame(data, idx, col)

        assert pd.testing.assert_frame_equal(df, test_df, check_dtype=False) is None

    ################################
    ### TEST GET LAST FIDELITIES ###
    ################################
    def test_get_last_fidelities(self, controller):
        """Test that the ``get_last_fidelities()`` method, gets the correct parameters."""
        for i, node in controller.node_sequence.items():
            node.output_parameters = {
                "platform_parameters": [(f"test_bus_{i}", 0, "param", 0), (f"test_bus_{i}", 1, "param", 1)],
                "fidelities": [(0, f"param_{i}", 1), (1, f"param_{i}", 0.967)],
            }
            node.previous_timestamp = 1999

        df = controller.get_last_fidelities()

        # Create the pandas DataFrame to test
        data = [
            [1, "zeroth_q0q1", datetime.fromtimestamp(1999)],
            [0.967, "zeroth_q0q1", datetime.fromtimestamp(1999)],
            [1, "first_q0", datetime.fromtimestamp(1999)],
            [0.967, "first_q0", datetime.fromtimestamp(1999)],
            [1, "second_q0", datetime.fromtimestamp(1999)],
            [0.967, "second_q0", datetime.fromtimestamp(1999)],
            [1, "third_q0", datetime.fromtimestamp(1999)],
            [0.967, "third_q0", datetime.fromtimestamp(1999)],
            [1, "fourth", datetime.fromtimestamp(1999)],
            [0.967, "fourth", datetime.fromtimestamp(1999)],
        ]
        idx = pd.MultiIndex.from_product(
            [
                [
                    "param_zeroth_q0q1",
                    "param_first_q0",
                    "param_second_q0",
                    "param_third_q0",
                    "param_fourth",
                ],
                [0, 1],
            ],
            names=["fidelity", "qubit"],
        )
        col = ["fidelity", "node_id", "datetime"]
        test_df = pd.DataFrame(data, idx, col)

        assert pd.testing.assert_frame_equal(df, test_df, check_dtype=False) is None

    ################################
    ##### TEST GET QUBITS TABLE ####
    ################################
    def test_get_qubits_table_and_test_create_empty_dataframe(self, controller):
        """Test that the ``get_qubits_table()`` and ``_create_empty_dataframes()`` methods, gets the correct structure and parameters."""
        for ind, (_, node) in enumerate(controller.node_sequence.items()):
            if node.node_id == "zeroth_q0q1":
                node.output_parameters = {
                    "platform_parameters": [(f"param_{ind}", 1, "test_bus", "0-1")],
                    "fidelities": [("0-1", f"fidelity_{ind}", 0.967)],
                }

            elif node.node_id == "fourth":
                node.node_id = "fourth_q1"
                node.output_parameters = {
                    "platform_parameters": [(f"param_{ind}", 1, "test_bus", 1)],
                    "fidelities": [(1, f"fidelity_{ind}", 0.967)],
                }

            else:
                node.output_parameters = {
                    "platform_parameters": [(f"param_{ind}", 1, "test_bus", 0)],
                    "fidelities": [(0, f"fidelity_{ind}", 0.967)],
                }
            node.previous_timestamp = 1999

        # Generate the empty dataframes, and the filled ones, for testing both functions.
        empty_q1_df, empty_q2_df = controller._create_empty_dataframes()
        q1_df, q2_df = controller.get_qubits_tables()

        # Create the pandas DataFrames idx, columns and data to test:
        idx = ["0-1", "0", "1"]
        data = [
            [1, 0.967],
            [1, 1, 1, "-", 0.967, 0.967, 0.967, "-"],
            ["-", "-", "-", 1, "-", "-", "-", 0.967],
        ]
        empty_data = [
            ["-", "-"],
            ["-", "-", "-", "-", "-", "-", "-", "-"],
            ["-", "-", "-", "-", "-", "-", "-", "-"],
        ]
        col_q1 = [
            "param_1_test_bus",
            "param_2_test_bus",
            "param_3_test_bus",
            "param_4_test_bus",
            "fidelity_1",
            "fidelity_2",
            "fidelity_3",
            "fidelity_4",
        ]
        col_q2 = [
            "param_0_test_bus",
            "fidelity_0",
        ]

        # Create the check empty dataframes:
        test_empty_q1_df = pd.DataFrame(empty_data[1:], idx[1:], col_q1)
        test_empty_q2_df = pd.DataFrame(empty_data[:1], idx[:1], col_q2)

        # Create the check filled dataframes:
        test_q1_df = pd.DataFrame(data[1:], idx[1:], col_q1)
        test_q2_df = pd.DataFrame(data[:1], idx[:1], col_q2)

        # Give name to the index to the test dataframes:
        for df in [test_q1_df, test_q2_df, test_empty_q1_df, test_empty_q2_df]:
            df.index.name = "qubit"

        # Testing the empty dataframes structure:
        assert (
            pd.testing.assert_frame_equal(
                empty_q1_df,
                test_empty_q1_df,
                check_dtype=False,
                check_like=True,
                check_index_type=False,
                check_column_type=False,
            )
            is None
        )
        assert (
            pd.testing.assert_frame_equal(
                empty_q2_df,
                test_empty_q2_df,
                check_dtype=False,
                check_like=True,
                check_index_type=False,
                check_column_type=False,
            )
            is None
        )

        # Testing that the dataframes got filled correctly
        assert (
            pd.testing.assert_frame_equal(
                q1_df, test_q1_df, check_dtype=False, check_like=True, check_index_type=False, check_column_type=False
            )
            is None
        )
        assert (
            pd.testing.assert_frame_equal(
                q2_df, test_q2_df, check_dtype=False, check_like=True, check_index_type=False, check_column_type=False
            )
            is None
        )

        # Undo the previous change to the node_id, for the nexts tests:
        for node in controller.node_sequence.values():
            if node.node_id == "fourth_q1":
                node.node_id = "fourth"

    def test_reorder_fidelities(self, controller):
        """Test that the reorder method, puts the fidelities in the front."""
        idx = ["0", "1"]

        empty_data = [
            ["-", "-", "-", "-", "-", "-", "-", "-"],
            ["-", "-", "-", "-", "-", "-", "-", "-"],
        ]
        col = [
            "param_1_test_bus",
            "param_2_test_bus",
            "param_3_test_bus",
            "param_4_test_bus",
            "fidelity_1",
            "fidelity_2",
            "fidelity_3",
            "fidelity_4",
        ]

        reordered_col = [
            "fidelity_4",
            "fidelity_3",
            "fidelity_2",
            "fidelity_1",
            "param_1_test_bus",
            "param_2_test_bus",
            "param_3_test_bus",
            "param_4_test_bus",
        ]

        df = pd.DataFrame(empty_data, idx, col)
        df = controller._reorder_fidelities(df)

        ordered_df = pd.DataFrame(empty_data, idx, reordered_col)

        assert pd.testing.assert_frame_equal(df, ordered_df, check_dtype=False, check_index_type=False) is None

    #########################
    ### TEST DEPENDENCIES ###
    #########################
    def test_dependencies(self, controller):
        """Test that dependencies return the correct dependencies."""
        result = controller._dependencies(nodes["zeroth_q0q1"])
        assert result == []

        result = controller._dependencies(nodes["fourth"])
        if controller.calibration_graph in [G0, G1]:
            assert third in result and second in result
            assert len(result) == 2

        elif controller.calibration_graph in [G2, G3, G4, G5]:
            assert third in result
            assert len(result) == 1

        elif controller.calibration_graph in [G6, G7]:
            assert third in result and first in result
            assert len(result) == 2

        elif controller.calibration_graph in [G8, G9]:
            assert third in result and second in result and first in result
            assert len(result) == 3

    def test_get_qubit_from_node(self, controller):
        """Test that the method ``get_qubit_from_node()`` returns the correct qubit index."""
        for node in controller.node_sequence.values():
            if isinstance(node.qubit_index, int):
                assert controller._get_qubit_from_node(node) == str(node.qubit_index)
            elif isinstance(node.qubit_index, list):
                assert controller._get_qubit_from_node(node) == "-".join(map(str, node.qubit_index))

    def test_get_bus_name_from_alias(self, controller):
        """Test that the method ``get_bus_name_from_alias()`` returns the correct bus name."""
        assert controller._get_bus_name_from_alias("test_bus_h56_gt") == "test_bus_gt"




class TestStaticMethodsFromCalibrationController:
    """Test that the static methods of ``CalibrationController`` behave well."""

    ###############################
    ### TEST IS TIMEOUT EXPIRED ###
    ###############################
    @pytest.mark.parametrize(
        "timestamp",
        [datetime(2023, 1, 1).timestamp(), datetime.now().timestamp() - 3600, datetime.now().timestamp() - 3600 * 23],
        ids=["timestampA", "timestampB", "timestampC"],
    )
    def test_timeout_expired(self, timestamp):
        """Tests cases where timeout should be expired."""
        timeout = 1800
        assert CalibrationController._is_timeout_expired(timestamp, timeout) is True

    @pytest.mark.parametrize(
        "timestamp",
        [datetime.now().timestamp(), datetime.now().timestamp() - 1700],
        ids=["timestampA", "timestampB"],
    )
    def test_timeout_not_expired(self, timestamp):
        """Test cases where timeout should not be expired."""
        timeout = 1800
        assert CalibrationController._is_timeout_expired(timestamp, timeout) is False



import numpy as np
import pytest
from qibo import gates

from qililab.digital.gate_decompositions import GateDecompositions, native_gates, translate_gates
from qililab.digital.native_gates import Drag, Wait


@pytest.fixture(name="test_gates")
def get_gates() -> list[gates.Gate]:
    """Fixture that returns a set of gates for the test"""
    return [gates.X(0), gates.Align(0, 16), gates.CNOT(0, 1), gates.RY(0, 2.5)]


def test_gatedecompositions(test_gates: list[gates.Gate]):
    """Test that class GateDecompositions creates a dictionary of gate translations

    Args:
        test_gates (list[gates.Gate]): list of gates to be translated
    """
    decomp = GateDecompositions()
    # add some fake decompositions
    decomp.add(gates.RY, lambda gate: [gates.U3(0, gate.parameters[0], 2, 0)])
    decomp.add(gates.X, [gates.Y(0), gates.Z(0)])
    decomp.add(gates.CNOT, lambda gate: [gates.SWAP(0, 1)])
    decomp.add(gates.Align, lambda gate: [Wait(0, gate.parameters[0])])

    # decomposed gates from fixture should look like this
    decomposed_gates = [gates.Y(0), gates.Z(0), Wait(0, 16), gates.SWAP(0, 1), gates.U3(0, 2.5, 2, 0)]

    assert isinstance(decomp.decompositions, dict)

    # check that decomposed gates is the same as decomposed_gates
    new_gates = []
    for gate in test_gates:
        new_gates.extend(decomp(gate))
    for gate, dec_gate in zip(new_gates, decomposed_gates):
        assert isinstance(gate, gates.Gate)
        assert gate.name == dec_gate.name
        assert gate.parameters == dec_gate.parameters


def test_translate_gates(test_gates: list[gates.Gate]):
    """Test that translate gates outputs a list of gates as expected

    Args:
        test_gates (gates.Gate): list of gates to be translated
    """

    tr_gates = translate_gates(test_gates)
    # decomposed gates from fixture should look like this
    decomposed_gates = [
        Drag(0, np.pi, 0),
        Wait(0, 16),
        Drag(1, np.pi / 2, -np.pi / 2),
        gates.RZ(1, np.pi),
        gates.CZ(0, 1),
        Drag(1, np.pi / 2, -np.pi / 2),
        gates.RZ(1, np.pi),
        Drag(0, 2.5, np.pi / 2),
    ]

    assert isinstance(tr_gates, list)
    for gate, dec_gate in zip(tr_gates, decomposed_gates):
        assert isinstance(gate, gates.Gate)
        assert gate.name == dec_gate.name
        assert gate.parameters == dec_gate.parameters


def test_native_gates():
    """Test native gates output is the intended set of gates"""

    assert native_gates() == (Drag, gates.CZ, Wait)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

""" Acquisition Result """

from dataclasses import dataclass, field

import numpy as np
import numpy.typing as npt
import pandas as pd

from qililab.constants import RESULTSDATAFRAME


@dataclass
class Acquisition:
    """Acquisition normalized
    Args:
        pulse_length (int): Duration (in ns) of the pulse.
        i_values: (NDArray[numpy.float32]): I data normalized
        q_values: (NDArray[numpy.float32]): I data normalize
        amplitude_values: (NDArray[numpy.float32]): amplitude values from I/Q normalized
        phase_values: (NDArray[numpy.float32]): phase values from I/Q normalized

    """

    integration_length: int
    i_values: npt.NDArray[np.float32]
    q_values: npt.NDArray[np.float32]
    amplitude_values: npt.NDArray[np.float32] = field(init=False)
    phase_values: npt.NDArray[np.float32] = field(init=False)
    acquisition: pd.DataFrame = field(init=False)
    data_dataframe_indices: set = field(init=False, default_factory=set)

    def __post_init__(self):
        """Create acquisitions"""
        self.i_values = self._normalized_data(data=self.i_values)
        self.q_values = self._normalized_data(data=self.q_values)
        self.amplitude_values = self._amplitudes(i_normalized=self.i_values, q_normalized=self.q_values)
        self.phase_values = self._phases(i_normalized=self.i_values, q_normalized=self.q_values)
        self.acquisition = self._create_acquisition()
        self.data_dataframe_indices = {
            RESULTSDATAFRAME.I,
            RESULTSDATAFRAME.Q,
            RESULTSDATAFRAME.AMPLITUDE,
            RESULTSDATAFRAME.PHASE,
        }

    def _create_acquisition(self) -> pd.DataFrame:
        """Transposes each of the acquired results arrays so that we have for each value a structure with i, q,
        amplitude, phase.
        For multiple values you may need to redefine this method.
        """

        return pd.DataFrame(
            {
                RESULTSDATAFRAME.I: self.i_values,
                RESULTSDATAFRAME.Q: self.q_values,
                RESULTSDATAFRAME.AMPLITUDE: self.amplitude_values,
                RESULTSDATAFRAME.PHASE: self.phase_values,
            }
        )

    def _normalized_data(self, data: npt.NDArray):
        """Normalizes the given data with the integration length,
        which should be the same as the pulse length.

        Args:
            data (list[float]): I or Q data from acquisition.

        Returns:
            NDArray[flaot]: Normalized data
        """
        return np.array(data) / self.integration_length

    def _amplitudes(self, i_normalized: npt.NDArray[np.float32], q_normalized: npt.NDArray[np.float32]):
        """Computes the amplitudes of a given I and Q data

        Args:
            i_normalized: (NDArray[numpy.float32]): I data normalized
            q_normalized: (NDArray[numpy.float32]): I data normalized

        Returns:
            NDArray[numpy.float]: amplitude
        """
        return 20 * np.log10(np.sqrt(i_normalized**2 + q_normalized**2))

    def _phases(self, i_normalized: npt.NDArray[np.float32], q_normalized: npt.NDArray[np.float32]):
        """Computes the phases of a given I and Q data

        Args:
            i_normalized: (NDArray[numpy.float32]): I data normalized
            q_normalized: (NDArray[numpy.float32]): I data normalized

        Returns:
            NDArray[numpy.float]: amplitude
        """
        return np.arctan2(q_normalized, i_normalized)



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""VNA Result class."""

from dataclasses import dataclass

import numpy as np
import numpy.typing as npt
import pandas as pd

from qililab.result.result import Result
from qililab.typings.enums import ResultName
from qililab.utils.factory import Factory


@Factory.register
@dataclass
class VNAResult(Result):  # TODO: Remove this class (it is useless)
    """VNAResult class."""

    data: npt.NDArray[np.float32]
    name: ResultName = ResultName.VECTOR_NETWORK_ANALYZER  # type: ignore[misc]

    def acquisitions(self) -> np.ndarray:
        """Return acquisition values."""
        return self.data

    def probabilities(self) -> pd.DataFrame:
        """Return probabilities of being in the ground and excited state.

        Returns:
            tuple[float, float]: Probabilities of being in the ground and excited state.
        """
        raise NotImplementedError

    @property
    def array(self) -> np.ndarray:
        return np.array([])

    def to_dict(self) -> dict:
        return {}



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""This module contains the waveforms that can be played within a QProgram.

.. currentmodule:: qililab

Waveforms
~~~~~~~~~~

.. autosummary::
    :toctree: api

    ~Waveform
    ~Square
    ~Ramp
    ~Gaussian
    ~Arbitrary
    ~DragCorrection
    ~Chained
    ~IQPair
    ~FlatTop
"""

from .arbitrary import Arbitrary
from .chained import Chained
from .drag_correction import DragCorrection
from .flat_top import FlatTop
from .gaussian import Gaussian
from .iq_pair import IQPair
from .ramp import Ramp
from .square import Square
from .waveform import Waveform

__all__ = ["Arbitrary", "Chained", "DragCorrection", "FlatTop", "Gaussian", "IQPair", "Ramp", "Square", "Waveform"]



"""Tests for the pulse distortion class."""
import itertools
import re

import numpy as np
import pytest

from qililab.pulse import Pulse
from qililab.pulse.pulse_distortion import BiasTeeCorrection, ExponentialCorrection, LFilterCorrection
from qililab.pulse.pulse_distortion.pulse_distortion import PulseDistortion
from qililab.pulse.pulse_shape import Cosine, Drag, Gaussian, Rectangular
from qililab.utils import Factory

# Parameters for the different corrections.
NORM_FACTOR = [0.95]
TAU_BIAS_TEE = [0.6]
TAU_EXPONENTIAL = [0.7]
AMP = [-2.1, 0.8]
A = [[0.7, 1.3]]
B = [[0.5, 0.6]]

# Parameters of the Pulse and its envelope.
AMPLITUDE = [-0.8, 0.9, 0]
PHASE = [0, np.pi / 3]
DURATION = [50, 37]  # Since we don't have the SNZ in this test, we can try an odd duration here.
FREQUENCY = [0.7e9]
RESOLUTION = [1.0]
SHAPE = [
    Rectangular(),
    Cosine(),
    Cosine(lambda_2=0.3),
    Gaussian(num_sigmas=4),
    Drag(num_sigmas=4, drag_coefficient=1.0),
]


@pytest.fixture(
    name="pulse_distortion",
    params=[
        ExponentialCorrection(tau_exponential=tau_exponential, amp=amp, norm_factor=norm_factor)
        for tau_exponential, amp, norm_factor in itertools.product(TAU_EXPONENTIAL, AMP, NORM_FACTOR)
    ]
    + [BiasTeeCorrection(tau_bias_tee=tau_bias_tee) for tau_bias_tee in TAU_BIAS_TEE]
    + [
        LFilterCorrection(a=a, b=b, norm_factor=norm_factor)
        for a, b, norm_factor in itertools.product(A, B, NORM_FACTOR)
    ],
)
def fixture_pulse_distortion(request: pytest.FixtureRequest) -> ExponentialCorrection:
    """Fixture for the pulse distortion class."""
    return request.param


@pytest.fixture(
    name="envelope",
    params=[
        Pulse(amplitude=amplitude, phase=phase, duration=duration, frequency=frequency, pulse_shape=shape).envelope(
            resolution=resolution
        )
        for amplitude, phase, duration, frequency, resolution, shape in itertools.product(
            AMPLITUDE, PHASE, DURATION, FREQUENCY, RESOLUTION, SHAPE
        )
    ],
)
def fixture_envelope(request: pytest.FixtureRequest) -> np.ndarray:
    """Fixture for an envelope."""
    return request.param


def return_corrected_envelopes_examples(
    pulse_distortion: PulseDistortion, envelope: np.ndarray, norm_factors: list[float]
):
    """Helper function that returns examples of envelopes with & without auto_norm"""
    norm_corr_envelopes = [pulse_distortion.apply(envelope=envelope)]
    norm_corr_envelopes.append(
        ExponentialCorrection(tau_exponential=1.3, amp=2.0, norm_factor=norm_factors[0]).apply(
            envelope=norm_corr_envelopes[0]
        )
    )
    norm_corr_envelopes.append(
        BiasTeeCorrection(tau_bias_tee=0.5, norm_factor=norm_factors[1]).apply(envelope=norm_corr_envelopes[1])
    )
    not_norm_corr_envelopes = [
        ExponentialCorrection(tau_exponential=0.5, amp=-5.0, auto_norm=False, norm_factor=norm_factors[1]).apply(
            envelope=norm_corr_envelopes[1]
        )
    ]
    not_norm_corr_envelopes.append(
        LFilterCorrection(a=[0.7, 1.3], b=[0.5, 0.6], norm_factor=norm_factors[0]).apply(
            envelope=not_norm_corr_envelopes[0]
        )
    )
    return norm_corr_envelopes, not_norm_corr_envelopes


class TestPulseDistortion:
    """Unit tests checking the PulseDistortion attributes and methods"""

    def test_apply_method(self, pulse_distortion: PulseDistortion, envelope: np.ndarray):
        """Test for the apply method."""
        norm_factors = [0.8, 1.15]
        norm_corr_envelopes, not_norm_corr_envelopes = return_corrected_envelopes_examples(
            pulse_distortion, envelope, norm_factors
        )

        # Basic checks
        for corr_envelope in norm_corr_envelopes + not_norm_corr_envelopes:
            assert corr_envelope is not None
            assert isinstance(envelope, np.ndarray)
            assert len(envelope) == len(corr_envelope)
            assert (
                not np.array_equal(envelope, corr_envelope)
                # Discarting the ampltidue = 0 case
                or np.max(np.abs(np.real(envelope))) == np.max(np.abs(np.real(corr_envelope))) == 0.0
            )

        # Autonorm = True cases, with amplitudes and norm_factors <= 1, have to return an envelope between -1 and 1:
        for norm_corr_envelope in norm_corr_envelopes:
            assert np.max((np.real(norm_corr_envelope))) <= 1
            assert np.min((np.real(norm_corr_envelope))) >= -1

    def test_norm_factors_and_auto_norm_in_apply_method(self, pulse_distortion: PulseDistortion, envelope: np.ndarray):
        """Test for the apply method."""
        norm_factors = [1.8, 0.35, 0.8]
        norm_corr_envelopes, not_norm_corr_envelopes = return_corrected_envelopes_examples(
            pulse_distortion, envelope, norm_factors
        )

        # Both Auto-norm = True & False, and Norm factors > 1 included
        assert (
            0.0  # Testing/Discarting the amplitude = 0 cases, for both maxs and mins.
            == round(np.max(np.abs(np.real(envelope))), 13)
            == round(np.min(np.abs(np.real(envelope))), 13)
            == round(np.max(np.abs(np.real(norm_corr_envelopes[0]))), 13)
            == round(np.min(np.abs(np.real(norm_corr_envelopes[0]))), 13)
            == round(np.max(np.abs(np.real(norm_corr_envelopes[1]))), 13)
            == round(np.min(np.abs(np.real(norm_corr_envelopes[1]))), 13)
            == round(np.max(np.abs(np.real(norm_corr_envelopes[2]))), 13)
            == round(np.min(np.abs(np.real(norm_corr_envelopes[2]))), 13)
            == round(np.max(np.abs(np.real(not_norm_corr_envelopes[0]))), 13)
            == round(np.min(np.abs(np.real(not_norm_corr_envelopes[0]))), 13)
            == round(np.max(np.abs(np.real(not_norm_corr_envelopes[1]))), 13)
            == round(np.min(np.abs(np.real(not_norm_corr_envelopes[1]))), 13)
        ) or (
            # Actual testing that the norm_factors are working properly, the factors
            # should correspond to the ones in the function "return_corrected_envelopes_examples()"
            round(np.max(np.abs(np.real(norm_corr_envelopes[0]))), 13)
            == round(np.max(np.abs(np.real(norm_corr_envelopes[1]))) / norm_factors[0], 13)
            == round(np.max(np.abs(np.real(norm_corr_envelopes[2]))) / (norm_factors[0] * norm_factors[1]), 13)
            == round(np.max(np.abs(np.real(envelope))) * pulse_distortion.norm_factor, 13)
            # Testing that the auto_norm changes the norm from the previous
            != round(np.max(np.abs(np.real(not_norm_corr_envelopes[0]))) / (norm_factors[0] * norm_factors[1]), 2)
            # Then we get back to a case with auto_norm = True, so we can check it again.
            == round(np.max(np.abs(np.real(not_norm_corr_envelopes[1]))) / (norm_factors[0] ** 2 * norm_factors[1]), 2)
        )

    def test_from_dict(self, pulse_distortion: PulseDistortion):
        """Test for the to_dict method."""
        dictionary = pulse_distortion.to_dict()
        pulse_distortion2 = PulseDistortion.from_dict(dictionary)

        dictionary2 = pulse_distortion2.to_dict()
        pulse_distortion3 = PulseDistortion.from_dict(dictionary2)

        assert isinstance(pulse_distortion2, Factory.get(name=pulse_distortion.name))
        assert isinstance(pulse_distortion3, Factory.get(name=pulse_distortion2.name))
        assert pulse_distortion2 is not None and pulse_distortion3 is not None
        assert isinstance(pulse_distortion2, PulseDistortion) and isinstance(pulse_distortion3, PulseDistortion)
        assert pulse_distortion == pulse_distortion2 == pulse_distortion3

    def test_incorrect_from_dict(self):
        """Test for the from_dict method with incorrect dictionary."""
        with pytest.raises(ValueError, match=re.escape(f"Class: {BiasTeeCorrection.name.value} to instantiate, does not match the given dict name {LFilterCorrection.name.value}")):
            BiasTeeCorrection.from_dict({"name": LFilterCorrection.name.value})

    def test_to_dict(self, pulse_distortion: PulseDistortion):
        """Test for the to_dict method."""
        dictionary = pulse_distortion.to_dict()

        assert dictionary is not None
        assert isinstance(dictionary, dict)

        if isinstance(pulse_distortion, BiasTeeCorrection):
            assert dictionary == {
                "name": pulse_distortion.name.value,
                "tau_bias_tee": pulse_distortion.tau_bias_tee,
                "sampling_rate": pulse_distortion.sampling_rate,
                "norm_factor": pulse_distortion.norm_factor,
                "auto_norm": pulse_distortion.auto_norm,
            }

        if isinstance(pulse_distortion, ExponentialCorrection):
            assert dictionary == {
                "name": pulse_distortion.name.value,
                "tau_exponential": pulse_distortion.tau_exponential,
                "amp": pulse_distortion.amp,
                "sampling_rate": pulse_distortion.sampling_rate,
                "norm_factor": pulse_distortion.norm_factor,
                "auto_norm": pulse_distortion.auto_norm,
            }

        if isinstance(pulse_distortion, LFilterCorrection):
            assert dictionary == {
                "name": pulse_distortion.name.value,
                "a": pulse_distortion.a,
                "b": pulse_distortion.b,
                "norm_factor": pulse_distortion.norm_factor,
                "auto_norm": pulse_distortion.auto_norm,
            }

    # TESTING CORNER CASES:

    @pytest.mark.parametrize(
        "envelope",
        [
            Rectangular().envelope(amplitude=0, duration=DURATION[1]),
            Rectangular().envelope(amplitude=0, duration=DURATION[0]),
            Cosine().envelope(amplitude=0, duration=DURATION[0]),
            Cosine(lambda_2=0.3).envelope(amplitude=0, duration=DURATION[0]),
            Gaussian(num_sigmas=4).envelope(amplitude=0, duration=DURATION[0]),
            Drag(num_sigmas=4, drag_coefficient=1.0).envelope(amplitude=0, duration=DURATION[0]),
        ],
    )
    def test_envelope_with_amplitude_0(self, pulse_distortion, envelope):
        """Testing that the corner case amplitude = 0 works properly."""
        corr_envelope = pulse_distortion.apply(envelope)
        assert 0.0 == np.max(np.abs(np.real(corr_envelope))) == np.min(np.abs(np.real(corr_envelope)))

    BIG_AMPLITUDES = [2.0, 1.2, -2.0]

    @pytest.mark.parametrize(
        "envelope",
        [
            Rectangular().envelope(amplitude=BIG_AMPLITUDES[0], duration=DURATION[1]),
            Rectangular().envelope(amplitude=BIG_AMPLITUDES[1], duration=DURATION[0]),
            Rectangular().envelope(amplitude=BIG_AMPLITUDES[2], duration=DURATION[0]),
            Cosine().envelope(amplitude=BIG_AMPLITUDES[1], duration=DURATION[0]),
            Cosine().envelope(amplitude=BIG_AMPLITUDES[2], duration=DURATION[1]),
            Cosine(lambda_2=0.3).envelope(amplitude=BIG_AMPLITUDES[1], duration=DURATION[0]),
            Gaussian(num_sigmas=4).envelope(amplitude=BIG_AMPLITUDES[1], duration=DURATION[0]),
            Drag(num_sigmas=4, drag_coefficient=1.0).envelope(amplitude=BIG_AMPLITUDES[1], duration=DURATION[0]),
        ],
    )
    def test_envelope_with_amplitude_bigger_than_1(self, pulse_distortion, envelope):
        """Testing that the corner case amplitude > 1 works properly."""
        assert np.max(np.abs(np.real(envelope))) > 1 or np.min(np.abs(np.real(envelope))) < 1

        corr_envelope = pulse_distortion.apply(envelope)
        assert np.max(np.abs(np.real(corr_envelope))) > 1 or np.min(np.abs(np.real(corr_envelope))) < 1



"""Test for the VectorNetworkAnalyzer class."""
from unittest.mock import MagicMock, patch

from qililab.typings.instruments.vector_network_analyzer import VectorNetworkAnalyzerDriver


class TestVectorNetworkAnalyzerDriver:
    """Test for the driver class at typings"""

    @patch("qililab.typings.instruments.vector_network_analyzer.pyvisa.ResourceManager")
    def test_post_init_method(self, mock_resource_manager):
        """Test the postinit function from the driver"""
        mock_resource = MagicMock(name="mock_resource")
        mock_resource_manager.return_value.open_resource.return_value = mock_resource
        vna_driver = VectorNetworkAnalyzerDriver("foo", "bar")
        vna_driver.__post_init__()
        assert vna_driver.driver == mock_resource

    @patch("qililab.typings.instruments.vector_network_analyzer.pyvisa.ResourceManager")
    def test_initial_setup_method(self, mock_resource_manager):
        """Test the initial setup method of the driver"""
        mock_resource = MagicMock(name="mock_resource")
        mock_resource_manager.return_value.open_resource.return_value = mock_resource
        vna_driver = VectorNetworkAnalyzerDriver("foo", "bar")
        vna_driver.initial_setup()
        vna_driver.driver.write.assert_called()

    @patch("qililab.typings.instruments.vector_network_analyzer.pyvisa.ResourceManager")
    def test_reset_method(self, mock_resource_manager):
        """Test the reset method of the driver"""
        mock_resource = MagicMock(name="mock_resource")
        mock_resource_manager.return_value.open_resource.return_value = mock_resource
        vna_driver = VectorNetworkAnalyzerDriver("foo", "bar")
        vna_driver.reset()
        vna_driver.driver.write.assert_called_with("SYST:PRES; *OPC?")

    @patch("qililab.typings.instruments.vector_network_analyzer.pyvisa.ResourceManager")
    def test_send_command(self, mock_resource_manager):
        """Test the send command method of the driver"""
        mock_resource = MagicMock(name="mock_resource")
        mock_resource_manager.return_value.open_resource.return_value = mock_resource
        vna_driver = VectorNetworkAnalyzerDriver("foo", "bar")
        vna_driver.send_command("SENS1:AVER:COUN", "3")
        vna_driver.driver.write.assert_called_with("SENS1:AVER:COUN 3")

    @patch("qililab.typings.instruments.vector_network_analyzer.pyvisa.ResourceManager")
    def test_send_query_method(self, mock_resource_manager):
        """Test the send query method of the driver"""
        mock_resource = MagicMock(name="mock_resource")
        mock_resource_manager.return_value.open_resource.return_value = mock_resource
        vna_driver = VectorNetworkAnalyzerDriver("foo", "bar")
        vna_driver.send_query(":SENS1:SWE:MODE?")
        vna_driver.driver.query.assert_called_with(":SENS1:SWE:MODE?")

    @patch("qililab.typings.instruments.vector_network_analyzer.pyvisa.ResourceManager")
    def test_send_binary_query_method(self, mock_resource_manager):
        """Test the send binary query method of the driver"""
        mock_resource = MagicMock(name="mock_resource")
        mock_resource_manager.return_value.open_resource.return_value = mock_resource
        vna_driver = VectorNetworkAnalyzerDriver("foo", "bar")
        vna_driver.send_binary_query("CALC1:MEAS1:DATA:SDAT?")
        vna_driver.driver.query_binary_values.assert_called_with("CALC1:MEAS1:DATA:SDAT?")

    @patch("qililab.typings.instruments.vector_network_analyzer.pyvisa.ResourceManager")
    def test_set_timeout_method(self, mock_resource_manager):
        """Test the set timeout method of the driver"""
        mock_resource = MagicMock(name="mock_resource")
        mock_resource_manager.return_value.open_resource.return_value = mock_resource
        vna_driver = VectorNetworkAnalyzerDriver("foo", "bar")
        vna_driver.set_timeout(200)
        assert vna_driver.timeout == 200

    @patch("qililab.typings.instruments.vector_network_analyzer.pyvisa.ResourceManager")
    def test_read_method(self, mock_resource_manager):
        """Test the read method to directly from the device"""
        mock_resource = MagicMock(name="mock_resource")
        mock_resource_manager.return_value.open_resource.return_value = mock_resource
        vna_driver = VectorNetworkAnalyzerDriver("foo", "bar")
        vna_driver.read()
        vna_driver.driver.read.assert_called()

    @patch("qililab.typings.instruments.vector_network_analyzer.pyvisa.ResourceManager")
    def test_read_raw_method(self, mock_resource_manager):
        """Test the read method to directly from the device"""
        mock_resource = MagicMock(name="mock_resource")
        mock_resource_manager.return_value.open_resource.return_value = mock_resource
        vna_driver = VectorNetworkAnalyzerDriver("foo", "bar")
        vna_driver.read_raw()
        vna_driver.driver.read_raw.assert_called()


# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Utiliy class for hashing common classes."""
import hashlib

from qm import Program
from qm.serialization.generate_qua_script import generate_qua_script
from qpysequence import Sequence


def hash_qua_program(program: Program) -> str:
    """Hash a QUA program"""
    program_str = "\n".join(generate_qua_script(program).split("\n")[3:])
    return hashlib.md5(program_str.encode("utf-8"), usedforsecurity=False).hexdigest()


def hash_qpy_sequence(sequence: Sequence):
    """Hash a QPy Sequence"""
    sequence_str = repr(sequence)
    return hashlib.md5(sequence_str.encode("utf-8"), usedforsecurity=False).hexdigest()



"""This file tests the the ``qblox_d5a`` class"""
import pytest

from qililab.instruments.qblox import QbloxS4g


@pytest.fixture(name="s4g")
def fixture_s4g():
    """Fixture that returns an instance of a dummy QbloxD5a."""
    return QbloxS4g(
        {
            "alias": "s4g",
            "current": [],
            "span": [],
            "ramping_enabled": [],
            "ramp_rate": [],
            "dacs": [],
        }
    )


class TestQbloxS4g:
    """This class contains the unit tests for the ``qblox_d5a`` class."""



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Nested dataclass decorator."""
from dataclasses import dataclass, is_dataclass
from typing import get_type_hints


def nested_dataclass(*args, **kwargs):
    """Class decorator used to cast any dict/str attributes into its corresponding dataclass/enum classes."""

    def wrapper(cls):
        """Class wrapper."""
        cls = dataclass(cls, **kwargs)
        original_init = cls.__init__

        def __init__(self, *args, **kwargs):
            for name, value in kwargs.items():
                field_type = get_type_hints(cls).get(name, None)
                if is_dataclass(field_type):
                    new_obj = field_type(**value)
                    kwargs[name] = new_obj
            original_init(self, *args, **kwargs)

        cls.__init__ = __init__
        return cls

    return wrapper(args[0]) if args else wrapper



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""DragCorrection waveform."""

import numpy as np

from qililab.qprogram.decorators import requires_domain
from qililab.qprogram.variable import Domain
from qililab.yaml import yaml

from .gaussian import Gaussian
from .waveform import Waveform


@yaml.register_class
class DragCorrection(Waveform):
    """Calculates the first order drag correction of the imaginary (Ey) channel of a drive pulse. See https://arxiv.org/abs/0901.0534 (10).

    So far only implemented for Gaussian pulses.

    Args:
        drag_coefficient (float): drag coefficient
        waveform (Waveform): waveform on which the drag transformation is calculated
    """

    @requires_domain("drag_coefficient", Domain.Scalar)
    def __init__(self, drag_coefficient: float, waveform: Waveform):
        super().__init__()
        self.drag_coefficient = drag_coefficient
        self.waveform = waveform

    def envelope(self, resolution: int = 1) -> np.ndarray:
        """Returns the envelope corresponding to the drag correction.

        Args:
            resolution (int, optional): Pulse resolution. Defaults to 1.

        Returns:
            np.ndarray: Height of the envelope for each time step.
        """
        if isinstance(self.waveform, Gaussian):
            sigma = self.waveform.duration / self.waveform.num_sigmas
            mu = self.waveform.duration / 2
            x = np.arange(self.waveform.duration / resolution) * resolution

            return (-1 * self.drag_coefficient * (x - mu) / sigma**2) * self.waveform.envelope()
        raise NotImplementedError(f"Cannot apply drag correction on a {self.waveform.__class__.__name__} waveform.")

    def get_duration(self) -> int:
        """Get the duration of the waveform.

        Returns:
            int: The duration of the waveform in ns.
        """
        return self.waveform.get_duration()



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""dict_factory used for dataclass' asdict method."""
from enum import Enum


def dict_factory(data):
    """dict factory used in the asdict() dataclass function. Replace all Enum classes by its corresponding values
    and all BusElement objects with its corresponding settings dictionaries."""
    result: dict[str, list[dict[str, int | float | str]] | str | int | float] = {}
    for key, value in data:
        if isinstance(value, Enum):
            value = str(value.value)
        result = result | {key: value}
    return result



# Copyright 2023 Qilimanjaro Quantum Tech
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""CircuitRouter class"""

import contextlib

import networkx as nx
import numpy as np
from qibo import Circuit, gates
from qibo.transpiler.optimizer import Preprocessing
from qibo.transpiler.pipeline import Passes
from qibo.transpiler.placer import Placer, ReverseTraversal, StarConnectivityPlacer
from qibo.transpiler.router import Router, Sabre, StarConnectivityRouter

from qililab.config import logger
from qililab.digital.circuit_optimizer import CircuitOptimizer
from qililab.digital.native_gates import _GateHandler


class CircuitRouter:
    """Handles circuit routing, using a Placer and Router, and a coupling map. It has a single accessible method:

    - ``route(circuit: Circuit) -> tuple[Circuit, dict]``: Routes the virtual/logical qubits of a circuit, to the chip's physical qubits.

    Args:
        connectivity (nx.graph): Chip connectivity.
        placer (Placer | type[Placer] | tuple[type[Placer], dict], optional): ``Placer`` instance, or subclass ``type[Placer]`` to
            use, with optionally, its kwargs dict (other than connectivity), both in a tuple. Defaults to ``ReverseTraversal``.
        router (Router | type[Router] | tuple[type[Router], dict], optional): ``Router`` instance, or subclass ``type[Router]`` to
            use, with optionally, its kwargs dict (other than connectivity), both in a tuple. Defaults to ``Sabre``.
    """

    def __init__(
        self,
        connectivity: nx.graph,
        placer: Placer | type[Placer] | tuple[type[Placer], dict] | None = None,
        router: Router | type[Router] | tuple[type[Router], dict] | None = None,
    ):
        self.connectivity = connectivity
        """nx.graph: Chip connectivity."""

        self.preprocessing = Preprocessing(self.connectivity)
        """Preprocessing: Stage to add qubits in the original circuit to match the number of qubits in the chip."""

        self.router = self._build_router(router, self.connectivity)
        """Router: Routing stage, where the final_layout and swaps will be created. Defaults to Sabre."""

        self.placer = self._build_placer(placer, self.router, self.connectivity)
        """Placer: Layout stage, where the initial_layout will be created. Defaults to ReverseTraversal."""

        # Cannot use Star algorithms for non star-connectivities:
        if self._if_star_algorithms_for_nonstar_connectivity(self.connectivity, self.placer, self.router):
            raise (ValueError("StarConnectivity Placer and Router can only be used with star topologies"))

        # Transpilation pipeline passes:
        self.pipeline = Passes([self.preprocessing, self.placer, self.router], self.connectivity)
        """Routing pipeline passes: Preprocessing, Placer and Router passes. Defaults to Passes([Preprocessing, ReverseTraversal, Sabre])."""
        # 1) Preprocessing adds qubits in the original circuit to match the number of qubits in the chip.
        # 2) Routing stage, where the final_layout and swaps will be created.
        # 3) Layout stage, where the initial_layout will be created.

    def route(self, circuit: Circuit, iterations: int = 10) -> tuple[Circuit, list[int]]:
        """Routes the virtual/logical qubits of a circuit to the physical qubits of a chip. Returns and logs the final qubit layout.

        Check public docstring in :meth:`.CircuitTranspiler.route_circuit()` for more information.

        Args:
            circuit (Circuit): circuit to route.
            iterations (int, optional): Number of times to repeat the routing pipeline, to keep the best stochastic result. Defaults to 10.

        Returns:
            tuple [Circuit, list[int]: Routed circuit and its corresponding final layout (Initial Re-mapping + SWAPs routing) of
                the Original Logical Qubits (l_q) in the physical circuit (wires): [l_q in wire 0, l_q in wire 1, ...].

        Raises:
            ValueError: If StarConnectivity Placer and Router are used with non-star topologies.
            ValueError: If the final layout is not valid, i.e. a qubit is mapped to more than one physical qubit.
        """
        # Call the routing pipeline on the circuit, multiple times, and keep the best stochastic result:
        best_transp_circ, least_swaps, best_final_layout = self._iterate_routing(self.pipeline, circuit, iterations)

        if least_swaps is not None:
            logger.info(f"The best found routing, has {least_swaps} swaps.")
            logger.info(
                f"{best_transp_circ.wire_names}: Initial Re-mapping of the Original Logical Qubits (l_q), in the Physical Circuit: [l_q in wire 0, l_q in wire 1, ...]."
            )
            logger.info(
                f"{best_final_layout}: Final Re-mapping (Initial + SWAPs routing) of the Original Logical Qubits (l_q), in the Physical Circuit: [l_q in wire 0, l_q in wire 1, ...]."
            )
        else:
            logger.info("No routing was done. Most probably due to routing iterations being 0.")

        return best_transp_circ, best_final_layout

    @staticmethod
    def _iterate_routing(
        routing_pipeline: Passes, circuit: Circuit, iterations: int = 10
    ) -> tuple[Circuit, int | None, list[int]]:
        """Iterates through the routing pipeline to retain the best stochastic result. Returns and/or logs the final qubit layout.

        Args:
            routing_pipeline (Passes): Transpilation pipeline passes.
            circuit (Circuit): Circuit to route.
            iterations (int, optional): Number of times to repeat the routing pipeline, to keep the best stochastic result. Defaults to 10.

        Returns:
            tuple[Circuit, int | None, list[int]]: Best routed circuit, least number of swaps required, and the corresponding best final
                layout of the original logical qubits in the physical circuit: [Logical qubit in wire 1, Logical qubit in wire 2, ...].
        """
        # We repeat the routing pipeline a few times, to keep the best stochastic result:
        least_swaps: int | None = None
        for _ in range(iterations):
            # Call the routing pipeline on the circuit:
            transpiled_circ, final_layout = routing_pipeline(circuit)

            # Undo the initial remapping (wire_names), for executing in correct chips:
            transpiled_circ = CircuitRouter._apply_initial_remap(transpiled_circ)

            # Remove redundant swaps at the start of the transpiled circuit:
            transpiled_circ = CircuitOptimizer.remove_redundant_start_controlled_gates(transpiled_circ, gates.SWAP)

            # Get the number of swaps in the circuits:
            n_swaps = len(transpiled_circ.gates_of_type(gates.SWAP))

            # Checking which is the best transpilation:
            if least_swaps is None or n_swaps < least_swaps:
                least_swaps = n_swaps
                best_transpiled_circ, best_final_layout = transpiled_circ, final_layout

            # If a mapping needs no swaps, we are finished:
            if n_swaps == 0:
                break

        best_final_layout = CircuitRouter._get_logical_qubit_of_each_wire(best_final_layout)

        return best_transpiled_circ, least_swaps, best_final_layout

    @staticmethod
    def _apply_initial_remap(transpiled_circ: Circuit) -> Circuit:
        """Applies the initial remapping of the circuit (wire_names), to the qubits of all gates, so we can execute in the connected qubits.

        Args:
            transpiled_circ (Circuit): Circuit with the initial remapping.

        Returns:
            Circuit: Circuit with the initial remapping applied to the gate qubits.
        """
        new_queue = []
        for gate in transpiled_circ.queue:
            qubits = [transpiled_circ.wire_names.index(qubit) for qubit in gate.qubits]
            gate = _GateHandler.create_gate(type(gate).__name__, qubits, gate.init_kwargs)
            new_queue.append(gate)

        return _GateHandler.create_circuit_from_gates(new_queue, transpiled_circ.nqubits, transpiled_circ.wire_names)

    @staticmethod
    def _get_logical_qubit_of_each_wire(final_layout: dict[int, int]) -> list[int]:
        """Transforms Qibo's format for the final_layout into a permutation in a list.

        Args:
            final_layout (dict[int, int]): Qibo's final layout (Initial Re-mapping + SWAPs routing) of the
                Original Logical Qubits (l_q) in the physical circuit (wire): {l_q: wire where it ended for execution}.

        Returns:
            final_layout (list[int]): Final layout (Initial Re-mapping + SWAPs routing) of the Original Logical Qubits (l_q) in the Physical Circuit (wire), in a list.
                Each index is a physical qubit (wire), and its value the corresponding Original Logical Qubit (l_q) state its containing: [l_q in wire 0, l_q in wire 1, ...].
        """
        if CircuitRouter._if_layout_is_not_valid(final_layout):
            raise ValueError(
                f"The final layout: {final_layout} is not valid. i.e. a logical qubit is mapped to more than one physical qubit, or a key/value isn't a number. Try again, if the problem persists, try another placer/routing algorithm."
            )
        return [logical_q for logical_q, _ in sorted(final_layout.items(), key=lambda q_wire_in_1: q_wire_in_1[1])]

    @staticmethod
    def _if_star_algorithms_for_nonstar_connectivity(connectivity: nx.Graph, placer: Placer, router: Router) -> bool:
        """True if the StarConnectivity Placer or Router are being used without a star connectivity.

        Args:
            connectivity (nx.Graph): Chip connectivity.
            placer (Placer): Placer instance.
            router (Router): Router instance.

        Returns:
            bool: True if the StarConnectivity Placer or Router are being used without a star connectivity.
        """

        return not nx.is_isomorphic(connectivity, nx.star_graph(4)) and (
            isinstance(placer, StarConnectivityPlacer) or isinstance(router, StarConnectivityRouter)
        )

    @staticmethod
    def _highest_degree_node(connectivity: nx.Graph) -> int:
        """Returns the node with the highest degree in the connectivity graph.

        Args:
            connectivity (nx.Graph): Chip connectivity.

        Returns:
            int: Node with the highest degree in the connectivity graph.
        """
        return max(dict(connectivity.degree()).items(), key=lambda x: x[1])[0]

    @staticmethod
    def _if_layout_is_not_valid(layout: dict[int, int]) -> bool:
        """True if the layout is not valid.

        For example, if a qubit is mapped to more than one physical qubit. Or if the keys or values are not int.

        Args:
            layout (dict[int, int]): Final layout of the circuit.

        Returns:
            bool: True if the layout is not valid.
        """
        return (
            len(layout.values()) != len(set(layout.values()))
            or len(layout.keys()) != len(set(layout.keys()))
            or not all(np.issubdtype(type(value), np.integer) for value in layout.values())
            or not all(np.issubdtype(type(keys), np.integer) for keys in layout)
        )

    @staticmethod
    def _build_router(router: Router | type[Router] | tuple[type[Router], dict], connectivity: nx.Graph) -> Router:
        """Build a `Router` instance, given the pass router in whatever format and the connectivity graph.

        Args:
            router (Router | type[Router] | tuple[type[Router], dict]): Router instance, subclass or tuple(subclass, kwargs).
            connectivity (nx.Graph): Chip connectivity.

        Returns:
            Router: Router instance.

        Raises:
            ValueError: If the router is not a Router subclass or instance.
        """
        # If router is None, we build default one:
        if router is None:
            return Sabre(connectivity=connectivity)

        kwargs = {}
        if isinstance(router, tuple):
            router, kwargs = router

        # If the router is already an instance, we update the connectivity to the platform one:
        if isinstance(router, Router):
            if kwargs:
                logger.warning("Ignoring router kwargs, as the router is already an instance.")
            router.connectivity = connectivity
            logger.warning("Substituting the router connectivity by the transpiler/platform/coupling_map one.")
            return router

        # If the router is a Router subclass, we instantiate it:
        with contextlib.suppress(TypeError, ValueError):
            if issubclass(router, Router):
                return router(connectivity=connectivity, **kwargs)

        raise TypeError(
            f"`router` arg ({type(router)}), must be a `Router` instance, subclass or tuple(subclass, kwargs), in `execute()`, `compile()`, `transpile_circuit()` or `route_circuit()`."
        )

    def _build_placer(
        self, placer: Placer | type[Placer] | tuple[type[Placer], dict], router: Router, connectivity: nx.graph
    ) -> Placer:
        """Build a `Placer` instance, given the pass router in whatever format and the connectivity graph.

        Args:
            router (Placer | type[Placer] | tuple[type[Placer], dict]): Placer instance, subclass or tuple(subclass, kwargs).
            connectivity (nx.Graph): Chip connectivity.

        Returns:
            Placer: Placer instance.

        Raises:
            ValueError: If the router is not a Placer subclass or instance.
        """
        # If placer is None, we build default one:
        if placer is None:
            return ReverseTraversal(connectivity=connectivity, routing_algorithm=router)

        kwargs = {}
        if isinstance(placer, tuple):
            placer, kwargs = placer

        # For ReverseTraversal placer, we need to check the routing algorithm has correct connectivity:
        if placer == ReverseTraversal or isinstance(placer, ReverseTraversal):
            placer, kwargs = self._check_ReverseTraversal_routing_connectivity(placer, kwargs, connectivity, router)

        # If the placer is already an instance, we update the connectivity to the platform one:
        if isinstance(placer, Placer):
            if kwargs:
                logger.warning("Ignoring placer kwargs, as the placer is already an instance.")
            placer.connectivity = connectivity
            logger.warning("Substituting the placer connectivity by the transpiler/platform/coupling_map one.")
            return placer

        # If the placer is a Placer subclass, we instantiate it:
        with contextlib.suppress(TypeError, ValueError):
            if issubclass(placer, Placer):
                return placer(connectivity=connectivity, **kwargs)

        raise TypeError(
            f"`placer` arg ({type(placer)}), must be a `Placer` instance, subclass or tuple(subclass, kwargs), in `execute()`, `compile()`, `transpile_circuit()` or `route_circuit()`."
        )

    @staticmethod
    def _check_ReverseTraversal_routing_connectivity(
        placer: Placer | type[Placer], kwargs: dict, connectivity: nx.Graph, router: Router
    ) -> tuple[Placer | type[Placer], dict]:
        """Checks if the kwargs are valid for the Router subclass, of the ReverseTraversal Placer.

        Args:
            placer (Placer | type[Placer])
            placer_kwargs (dict): kwargs for the Placer subclass.
            connectivity (nx.Graph): Chip connectivity.
            router (Router): Original transpilation Router subclass.

        Raises:
            ValueError: If the routing_algorithm is not a Router subclass or instance.

        Returns:
            tuple[Placer | type[Placer], dict]: tuple containing the final placer and its kwargs
        """
        # If the placer is a ReverseTraversal instance, we update the connectivity to the platform one:
        if isinstance(placer, ReverseTraversal):
            placer.routing_algorithm.connectivity = connectivity
            logger.warning(
                "Substituting the ReverseTraversal router connectivity, by the transpiler/platform/coupling_map one."
            )
            return placer, kwargs

        # Else is placer is not an instance, we need to check the routing algorithm:

        # If no routing algorithm is passed, we use the Transpilation Router, for the ReverseTraversal too:
        if "routing_algorithm" not in kwargs:
            kwargs["routing_algorithm"] = router
            return placer, kwargs

        # If the routing algorithm is a Router instance, we update the connectivity to the platform one:
        if isinstance(kwargs["routing_algorithm"], Router):
            logger.warning(
                "Substituting the passed connectivity for the routing algorithm, by the platform/transpiler/coupling_map connectivity",
            )
            kwargs["routing_algorithm"].connectivity = connectivity
            return placer, kwargs

        # If the routing algorithm is a Router subclass, we instantiate it, with the platform connectivity:
        with contextlib.suppress(TypeError, ValueError):
            if issubclass(kwargs["routing_algorithm"], Router):
                kwargs["routing_algorithm"] = kwargs["routing_algorithm"](connectivity)
                return placer, kwargs

        # If the routing algorithm is not a Router subclass or instance, we raise an error:
        raise TypeError(
            f"`routing_algorithm` `Placer` kwarg ({kwargs['routing_algorithm']}) must be a `Router` subclass or instance, in `execute()`, `compile()`, `transpile_circuit()` or `route_circuit()`."
        )
